#!/usr/bin/env python3
"""
Detect devices present on the serial ports, that are attached via
FTDI USB to RS232 adaptors.  This will write a cache file that can
be used by the other utility scripts.

The script will recognise the following devices, and more than one device of
each type can be detected:

controller : Newport ESP300, MM4006 and SMC100
     lvpsu : Agilent E3647A, E3634A, Hameg (Rohde & Schwarz) HMP4040
     hvpsu : Keithley 2410, 2614b; ISEG SHQ 222M, 224M

The terminator should set to be the same on all devices to make sure each
device can cleanly reject commands intended for other devices. <CRLF> is
expected.
"""

import argparse
import collections
import concurrent.futures as cf
import functools
import itertools
import sys
import time

import serial
import serial.tools.list_ports as stlp

from mmcb import common
from mmcb import lexicon


##############################################################################
# command line option handler
##############################################################################


def check_arguments():
    """
    handle command line options

    --------------------------------------------------------------------------
    args : none
    --------------------------------------------------------------------------
    returns
        args : <class 'argparse.Namespace'>
            e.g. Namespace(assume=False, nocache=False)
    --------------------------------------------------------------------------
    """
    parser = argparse.ArgumentParser(
        description='A diagnostic tool to detect devices attached to a\
            computer\'s RS232 serial ports via FTDI USB to RS232 adaptors.\
            Devices that are recognised are:\
            Newport ESP300, MM4006 and SMC100 stage controllers;\
            Hameg (Rohde & Schwarz) HMP4040, Keithley 2410/2614B,\
            ISEG SHQ 222M/224M and Agilent E3647A/E3634A PSUs.\
            Writes a cache file containing detected devices that can be used\
            by other scripts in this suite.'
    )
    parser.add_argument(
        '--nocache',
        action='store_true',
        help='do not create cache file from list of detected devices',
    )

    return parser.parse_args()


##############################################################################
# detect devices on serial ports
##############################################################################


def find_controller(port, config):
    """
    Check if an ESP300 or mm4006 is present on the given serial port.

    Note that the device's error-FIFO is manually cleared to ensure further
    communications can be processed successfully.

    from reference/ESP300.pdf page 3-7 (page 47 of the PDF):

    "A controller command (or a sequence of commands) has to be terminated with
    a carriage return character. However, responses from the controller are
    always terminated by a carriage return/line feed combination. This setting
    may not be changed."

    Newport MM4006 responds to "VE" with:
    "ve mm4006 controller version  7.01 date 04-02-2003"
    mm4006 ignores rs232 commands when user is in the config menu.

    Note that this function is called quite late in the detection process. By
    the time it is called, this script will already have sent any esp300 or
    mm4006 present on the specified serial port multiple commands it would
    find incomprehensible, even if they were sent at a matching baud rate.
    Hence, the need to clear the ESP300's error FIFO before detection can
    commence.

    --------------------------------------------------------------------------
    args
        port : string
            port name
        config : dict
            serial port configuration
    --------------------------------------------------------------------------
    returns
        retval : tuple
            (category : string,
             settings : dict,
             (port : string,
              manufacturer : string,
              model : string,
              serial_number : string,
              detected : bool,
              channels : list,
              release_delay : float))
    --------------------------------------------------------------------------
    """
    retval = (None, None, (port, None, None, None, None, None, None))
    ser = serial.Serial()
    ser.apply_settings(config)
    ser.port = port

    # initialise and open serial port
    try:
        ser.open()
    except (OSError, serial.SerialException):
        sys.exit(f'could not open port {port}, exiting.')
    else:
        ser.reset_output_buffer()
        ser.reset_input_buffer()

        # manually empty ESP30X's 10 item error FIFO
        # the mm4006 - like the smc100 - only holds the last error
        #
        # FIXME
        #
        # The commands differ: TE? (esp300), TE (mm4006).
        # This does not seem to cause problems with the mm4006, however check
        # how the mm4006 handles this. The esp300 needs to have the FIFO
        # cleared; perhaps the mm4006 is more tolerant in its error handling
        # or it allows (undocumented) use of a trailing '?' for queries.
        #
        # Also investigate possibility of early termination to (slightly)
        # reduce detection time:
        #
        # ESP300.pdf, p. 3-139
        # TE? returns '0' if no error
        #
        # MM4K6_701_E3.pdf, p. 3.143
        # TE returns 'TE@' if no error
        esp300_read_error = lexicon.motion('esp300', 'read error')
        esp300_fifo_size = 10
        for _ in itertools.repeat(None, esp300_fifo_size):
            ser.write(esp300_read_error)
            try:
                ser.readline()
            except serial.SerialException:
                print(f'cannot access serial port {port}')
                ser.close()
                return retval

        # command string is identical for esp300 and mm4006
        ser.write(lexicon.motion('esp300', 'read controller version'))

        # get response
        try:
            controller = ser.readline().decode('utf-8').strip().lower()
        except UnicodeDecodeError:
            ser.close()
        else:
            channels = ['']
            release_delay = None

            # we might have an ESP300 or ESP301 controller/driver
            if 'esp30' in controller:
                manufacturer = 'newport'
                model = controller.partition(' ')[0]
                serial_number = ''
                detected = True
                settings = ser.get_settings()
                retval = (
                    'controller',
                    settings,
                    (
                        port,
                        manufacturer,
                        model,
                        serial_number,
                        detected,
                        channels,
                        release_delay,
                    ),
                )
            elif 'mm4006' in controller:
                manufacturer = 'newport'
                model = controller.split()[1]
                serial_number = ''
                detected = True
                settings = ser.get_settings()
                retval = (
                    'controller',
                    settings,
                    (
                        port,
                        manufacturer,
                        model,
                        serial_number,
                        detected,
                        channels,
                        release_delay,
                    ),
                )

            # no device detected
            ser.close()

    return retval


def find_controller_smc(port, config):
    """
    Check if an SMC100 controller is present on the given serial port.

    The SMC100 controller has non-configurable RS232 settings so handle it
    separately in this function rather than making find_controller()
    overly long.

    Note that the device's error-FIFO is manually cleared to ensure further
    communications can be processed successfully.

    Newport SMC100 responds to 1VE with:
    "1VE SMC_CC - Controller-driver version  3. 1. 2"

    ASSUME that we only have one SMC controller, and it is at address 1.

    --------------------------------------------------------------------------
    args
        port : string
            port name
        config : dict
            serial port configuration
    --------------------------------------------------------------------------
    returns
        retval : tuple
            (category : string,
             settings : dict,
             (port : string,
              manufacturer : string,
              model : string,
              serial_number : string,
              detected : bool,
              channels : list,
              release_delay : float))
    --------------------------------------------------------------------------
    """
    retval = (None, None, (port, None, None, None, None, None, None))
    ser = serial.Serial()
    ser.apply_settings(config)
    ser.port = port

    # initialise and open serial port
    try:
        ser.open()
    except (OSError, serial.SerialException):
        sys.exit(f'could not open port {port}, exiting.')
    else:
        ser.reset_output_buffer()
        ser.reset_input_buffer()

        # the SMC100 retains a record of the last error only
        ser.write(lexicon.motion('smc', 'read error', identifier='1'))
        try:
            ser.readline()
        except serial.SerialException:
            print(f'cannot access serial port {port}')
            ser.close()
            return retval

        ser.write(lexicon.motion('smc', 'read controller version', identifier='1'))

        # get response
        try:
            controller = ser.readline().decode('utf-8').strip().lower()
        except UnicodeDecodeError:
            ser.close()
        else:
            if 'smc' in controller:
                release_delay = None
                channels = ['']
                manufacturer = 'newport'
                model = 'smc'
                serial_number = ''
                detected = True
                settings = ser.get_settings()
                retval = (
                    'controller',
                    settings,
                    (
                        port,
                        manufacturer,
                        model,
                        serial_number,
                        detected,
                        channels,
                        release_delay,
                    ),
                )

            # no device detected
            ser.close()

    return retval


def find_lvpsu(port, config):
    """
    check if any of these devices are present on the given serial port:

           Keysight/Agilent E3647A (dual output)
    Hewlett-Packard/Agilent E3634A (single output)

    notes for this type of PSU:

    a null-modem/crossover adaptor needs to be added to the serial cable

    test commands using miniterm rather than putty, e.g.

    [avt@pc048057 dev]$ miniterm.py --port=/dev/ttyUSB1
    --- Miniterm on /dev/ttyUSB1: 9600,8,N,1 ---
    --- Quit: Ctrl+]  |  Menu: Ctrl+T | Help: Ctrl+T followed by Ctrl+H ---
    Agilent Technologies,E3647A,0,1.7-5.0-1.0

    When using Keysight Connection Expert 2018 on Windows, the Flow control
    setting in Windows Device Manger seems to have no effect. Hardware, None,
    or Xon / Xoff all seem to work.

    in essence, ignore what the manual says about remote mode and dsrstr:

    (1) do not put the PSU into remote mode. Keysight Connection Expert
    2018 doesn't do this on Windows, and it seems to cause problems here
    (2) do not use dsrdtr

    response for Keysight E3634A:

    [avt@pc048057 utilities]$ miniterm.py --port=/dev/ttyUSB0
    --- Miniterm on /dev/ttyUSB0: 9600,8,N,1 ---
    --- Quit: Ctrl+]  |  Menu: Ctrl+T | Help: Ctrl+T followed by Ctrl+H ---
    HEWLETT-PACKARD,E3634A,0,2.4-6.1-2.1

    SYST:VERS? yields the SCPI version e.g. 1997.0
    YYYY.V where the Y represents the year of the version,
    and the V represents the version number for that year.

    --------------------------------------------------------------------------
    args
        port : string
            port name
        config : dict
            serial port configuration
    --------------------------------------------------------------------------
    returns
        retval : tuple
            (category : string,
             settings : dict,
             (port : string,
              manufacturer : string,
              model : string,
              serial_number : string,
              detected : bool,
              channels : list,
              release_delay : float))
    --------------------------------------------------------------------------
    """
    retval = (None, None, (port, None, None, None, None, None, None))
    ser = serial.Serial()
    ser.apply_settings(config)
    ser.port = port

    # initialise and open serial port
    try:
        ser.open()
    except (OSError, serial.SerialException):
        sys.exit(f'could not open port {port}, exiting.')
    else:
        ser.reset_input_buffer()
        ser.reset_output_buffer()

        # ask device to identify itself
        ser.write(lexicon.power('e3647a', 'identify'))

        # attempt to read back response
        try:
            response = ser.readline()
        except (serial.SerialException, AttributeError):
            print(f'cannot access serial port {port}')
        else:
            response = response.lower()
            if b'agilent' in response or b'hewlett-packard' in response:
                parts = response.decode('utf-8').split(',')
                manufacturer = 'agilent'
                model = parts[1].strip()
                serial_number = ''
                detected = True
                settings = ser.get_settings()

                if model in {'e3647a', 'e3634a'}:
                    if model == 'e3647a':
                        release_delay = 0.01
                        channels = ['1', '2']
                    else:
                        release_delay = None
                        channels = ['']

                    retval = (
                        'lvpsu',
                        settings,
                        (
                            port,
                            manufacturer,
                            model,
                            serial_number,
                            detected,
                            channels,
                            release_delay,
                        ),
                    )
                else:
                    print(f'lvpsu: unrecognised model {model}')

        ser.close()

    return retval


def find_lvpsu_hmp4040(port, config):
    """
    check if any of these devices are present on the given serial port:

    Rohde & Schwarz HMP4040

    response to *IDN?:

    macbook:utilities avt$ python3 -m serial.tools.miniterm

    --- Available ports:
    ---  1: /dev/cu.Bluetooth-Incoming-Port 'n/a'
    ---  2: /dev/cu.SSDC         'n/a'
    ---  3: /dev/cu.usbserial-AH06DY15 'FT232R USB UART'
    --- Enter port index or full name: 3
    --- Miniterm on /dev/cu.usbserial-AH06DY15  9600,8,N,1 ---
    --- Quit: Ctrl+] | Menu: Ctrl+T | Help: Ctrl+T followed by Ctrl+H ---
    HAMEG,HMP4040,103781,HW50020001/SW2.51

    Some models also identify as ROHDE&SCHWARZ instead of HAMEG.

    (pve) pi@raspberrypi $ python3 -m serial.tools.miniterm

    --- Available ports:
    ---  1: /dev/ttyAMA0         'ttyAMA0'
    ---  2: /dev/ttyUSB0         'FT232R USB UART - FT232R USB UART'
    --- Enter port index or full name: 2
    --- Miniterm on /dev/ttyUSB0  9600,8,N,1 ---
    --- Quit: Ctrl+] | Menu: Ctrl+T | Help: Ctrl+T followed by Ctrl+H ---
    ROHDE&SCHWARZ,HMP4040,120224,HW50020003/SW2.70

    --------------------------------------------------------------------------
    args
        port : string
            port name
        config : dict
            serial port configuration
    --------------------------------------------------------------------------
    returns
        retval : tuple
            (category : string,
             settings : dict,
             (port : string,
              manufacturer : string,
              model : string,
              serial_number : string,
              detected : bool,
              channels : list,
              release_delay : float))
    --------------------------------------------------------------------------
    """
    retval = (None, None, (port, None, None, None, None, None, None))
    ser = serial.Serial()
    ser.apply_settings(config)
    ser.port = port

    # initialise and open serial port
    try:
        ser.open()
    except (OSError, serial.SerialException):
        sys.exit(f'could not open port {port}, exiting.')
    else:
        ser.reset_input_buffer()
        ser.reset_output_buffer()

        # ask device to identify itself
        ser.write(lexicon.power('hmp4040', 'identify'))

        # attempt to read back response
        try:
            response = ser.readline()
        except (serial.SerialException, AttributeError):
            print(f'cannot access serial port {port}')
        else:
            response = response.lower()
            if b'hameg' in response or b'rohde' in response:
                release_delay = 0.026
                parts = response.decode('utf-8').split(',')
                manufacturer = 'hameg'
                model = parts[1].strip()
                serial_number = parts[2].strip()
                detected = True
                settings = ser.get_settings()
                channels = ['1', '2', '3', '4']
                retval = (
                    'lvpsu',
                    settings,
                    (
                        port,
                        manufacturer,
                        model,
                        serial_number,
                        detected,
                        channels,
                        release_delay,
                    ),
                )

        ser.close()

    return retval


def find_hvpsu(port, config):
    """
    check if one of the following power supplies is present on the given
    serial port

        Keithley 2410
        ISEG SHQ 222M
        ISEG SHQ 224M

    this function also successfully identifies:
    Keithley 6517B electrometer / high resistance meter

    Keithley 2410 response to *IDN?
    KEITHLEY INSTRUMENTS INC.,MODEL 2410,4343654,C34 Sep 21 2016 15:30:00/A02  /K/M
    --------------------------------------------------------------------------
    args
        port : string
            port name
        config : dict
            serial port configuration
    --------------------------------------------------------------------------
    returns
        retval : tuple
            (category : string,
             settings : dict,
             (port : string,
              manufacturer : string,
              model : string,
              serial_number : string,
              detected : bool,
              channels : list,
              release_delay : float))
    --------------------------------------------------------------------------
    """
    retval = (None, None, (port, None, None, None, None, None, None))
    ser = serial.Serial()
    ser.apply_settings(config)
    ser.port = port

    # initialise and open serial port
    try:
        ser.open()
    except (OSError, serial.SerialException):
        sys.exit(f'could not open port {port}, exiting.')
    else:
        ser.reset_input_buffer()
        ser.reset_output_buffer()
        ser.write(lexicon.power('2410', 'identify'))

        # attempt to read back response
        try:
            response = ser.readline().lower()
        except serial.SerialException:
            print(f'cannot access serial port {port}')
        else:
            if b'keithley' in response:
                release_delay = None
                parts = response.decode('utf-8').split(',')
                manufacturer = parts[0].partition(' ')[0]
                model = parts[1].rpartition(' ')[-1]
                serial_number = parts[2]
                detected = True
                channels = ['']
                settings = ser.get_settings()
                retval = (
                    'hvpsu',
                    settings,
                    (
                        port,
                        manufacturer,
                        model,
                        serial_number,
                        detected,
                        channels,
                        release_delay,
                    ),
                )
            else:
                # check for ISEG SHQ dual-channel power supplies

                ser.reset_input_buffer()
                ser.reset_output_buffer()
                time.sleep(0.5)

                # send CRLF first for sync
                ser.write(lexicon.power('shq', 'synchronise'))
                ser.readline()  # consume local echo
                ser.readline()  # consume local echo
                ser.readline()  # consume response to command in buffer

                # '#' command: read device identifier
                # device identifier does not contain a model number:
                # serial number;software release;Vout max;Iout max
                # e.g. ISEG SHQ 224M returns '484216;3.09;4000V;3mA'
                # e.g. ISEG SHQ 222M returns '484230;3.10;2000V;6mA'
                ser.write(lexicon.power('shq', 'identify'))
                ser.readline()  # consume local echo
                response = ser.readline()
                parts = [x.strip() for x in response.decode('utf-8').split(';')]
                if len(parts) == 4:
                    release_delay = 0.01
                    manufacturer = 'iseg'
                    model = 'shq'
                    serial_number = parts[0]
                    detected = True
                    channels = ['1', '2']
                    settings = ser.get_settings()
                    retval = (
                        'hvpsu',
                        settings,
                        (
                            port,
                            manufacturer,
                            model,
                            serial_number,
                            detected,
                            channels,
                            release_delay,
                        ),
                    )

        ser.close()

    return retval


def find_hvpsu_2614b(port, config):
    """
    check if a Keithley 2614b is present on the given serial port

    note that the language used for this generation of Keithley is quite
    different from earlier models

    e.g.

    d=localnode.description
    print(d)

    returns:

    Keithley Instruments SMU 2614B - 4428182

    --------------------------------------------------------------------------
    args
        port : string
            port name
        config : dict
            serial port configuration
    --------------------------------------------------------------------------
    returns
        retval : tuple
            (category : string,
             settings : dict,
             (port : string,
              manufacturer : string,
              model : string,
              serial_number : string,
              detected : bool,
              channels : list,
              release_delay : float))
    --------------------------------------------------------------------------
    """
    retval = (None, None, (port, None, None, None, None, None, None))
    ser = serial.Serial()
    ser.apply_settings(config)
    ser.port = port

    # initialise and open serial port
    try:
        ser.open()
    except (OSError, serial.SerialException):
        sys.exit(f'could not open port {port}, exiting.')
    else:
        ser.reset_input_buffer()
        ser.reset_output_buffer()
        ser.write(lexicon.power('2614b', 'identify'))

        # attempt to read back response
        try:
            response = ser.readline().lower()
        except serial.SerialException:
            print(f'cannot access serial port {port}')
        else:
            if b'keithley' in response:
                try:
                    release_delay = 0.01
                    parts = response.decode('utf-8').split()
                    manufacturer = parts[0]
                    model = parts[3]
                    serial_number = parts[5]
                    detected = True
                    channels = ['a', 'b']
                    settings = ser.get_settings()
                    retval = (
                        'hvpsu',
                        settings,
                        (
                            port,
                            manufacturer,
                            model,
                            serial_number,
                            detected,
                            channels,
                            release_delay,
                        ),
                    )
                except IndexError:
                    pass

        ser.close()

    return retval


def find_dmm_6500(port, config):
    """
    check if a Keithley DMM6500 is present on the given serial port

    *IDN? returns:
    KEITHLEY INSTRUMENTS,MODEL DMM6500,04592428,1.7.12b

    --------------------------------------------------------------------------
    args
        port : string
            port name
        config : dict
            serial port configuration
    --------------------------------------------------------------------------
    returns
        retval : tuple
            (category : string,
             settings : dict,
             (port : string,
              manufacturer : string,
              model : string,
              serial_number : string,
              detected : bool,
              channels : list,
              release_delay : float))
    --------------------------------------------------------------------------
    """
    retval = (None, None, (port, None, None, None, None, None, None))
    ser = serial.Serial()
    ser.apply_settings(config)
    ser.port = port

    # initialise and open serial port
    try:
        ser.open()
    except (OSError, serial.SerialException):
        sys.exit(f'could not open port {port}, exiting.')
    else:
        ser.reset_input_buffer()
        ser.reset_output_buffer()
        ser.write(lexicon.instrument('dmm6500', 'identify'))

        # attempt to read back response
        try:
            response = ser.readline().lower()
        except serial.SerialException:
            print(f'cannot access serial port {port}')
        else:
            if b'keithley' in response:
                try:
                    release_delay = 0.01
                    # ['KEITHLEY INSTRUMENTS', 'MODEL DMM6500', '04592428', '1.7.12b']
                    parts = response.decode('utf-8').lower().split(',')
                    manufacturer = parts[0].split()[0]
                    model = parts[1].split()[-1]
                    serial_number = parts[2]
                    detected = True
                    channels = ['']
                    settings = ser.get_settings()
                    retval = (
                        'instrument',
                        settings,
                        (
                            port,
                            manufacturer,
                            model,
                            serial_number,
                            detected,
                            channels,
                            release_delay,
                        ),
                    )
                except IndexError:
                    pass

        ser.close()

    return retval


def detect_device_on_port(port, tests, configs):
    """
    try all tests on the given port in a bid to identify which category of
    device is present

    --------------------------------------------------------------------------
    args
        port : string
            port name
        tests : dict
            contains the function to call for each category of device
        configs : dict of dicts
            {device_type: serial_port_configuration, ...}
    --------------------------------------------------------------------------
    returns
        retval : tuple
            (category : string,
             settings : dict,
             (port : string,
              manufacturer : string,
              model : string,
              serial_number : string,
              detected : bool,
              channels : list,
              release_delay : float))
    --------------------------------------------------------------------------
    """
    retval = (None, None, (port, None, None, None, None, None, None))

    for test in tests:
        retval = tests[test](port, configs[test])
        if retval[0] is not None:
            break

        # macOS seems to process serial port events more swiftly than CentOS
        # and Windows 7: need to give the devices enough time to respond
        # between requests
        time.sleep(0.3)

    return retval


##############################################################################
# format detail for detected device
##############################################################################


def display_found(device_type, value, padding):
    """
    print details for a device identified on a serial port

    --------------------------------------------------------------------------
    args
        device_type : string
            e.g. 'hvpsu'
        value : tuple (string, string, string, string, bool)
            e.g. (port, manufacturer, model, serial_number, detected_flag)
        padding : int
            the length of the longest device type
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    port, manufacturer, model, serial_number, *_ = value

    if device_type is None:
        print(f'{"None".rjust(padding)} on {port}: (no device identified)')
    else:
        details = [d for d in (manufacturer, model) if d]
        if serial_number:
            details.append(f's.no. {serial_number}')

        suffix = ' '.join(details)
        print(f'{device_type.rjust(padding)} on {port}: {suffix}')


##############################################################################
# main
##############################################################################


def main():
    """
    detect devices present on the serial ports, that are attached via
    FTDI USB to RS232 adaptors
    """
    args = check_arguments()

    ports = [com.device for com in stlp.comports() if common.rs232_port_is_valid(com)]
    if not ports:
        all_ports = ', '.join(com.device for com in stlp.comports())
        sys.exit(f'No usable serial ports found: {all_ports}')

    tests = collections.OrderedDict(
        {
            'dmm_6500': find_dmm_6500,
            'lvpsu': find_lvpsu,
            'lvpsu_hmp4040': find_lvpsu_hmp4040,
            'hvpsu': find_hvpsu,
            'hvpsu_2614b': find_hvpsu_2614b,
            'controller': find_controller,
            'controller_smc': find_controller_smc,
        }
    )

    # define serial port connection settings for each device
    # see https://pyserial.readthedocs.io/en/latest/pyserial_api.html
    common_configuration = {
        'bytesize': serial.EIGHTBITS,
        'parity': serial.PARITY_NONE,
        'stopbits': serial.STOPBITS_ONE,
        'dsrdtr': False,
        'timeout': 1,
        'write_timeout': 1,
        'inter_byte_timeout': None,
    }
    configs = {
        'hvpsu': {
            'baudrate': 9600,
            'xonxoff': False,
            'rtscts': False,
        },
        'hvpsu_2614b': {
            'baudrate': 57600,
            'xonxoff': False,
            'rtscts': False,
        },
        'dmm_6500': {
            'baudrate': 38400,
            'xonxoff': False,
            'rtscts': False,
        },
        'controller': {
            'baudrate': 19200,
            'xonxoff': False,
            'rtscts': False,
        },
        'controller_smc': {
            'baudrate': 57600,
            'xonxoff': True,
            'rtscts': False,
        },
        'lvpsu': {
            'baudrate': 9600,
            'xonxoff': False,
            'rtscts': False,
        },
        'lvpsu_hmp4040': {
            'baudrate': 9600,
            'xonxoff': False,
            'rtscts': True,
        },
    }
    # create unions of common and device-specific configurations
    configs = {
        device: {**device_specific_configuration, **common_configuration}
        for device, device_specific_configuration in configs.items()
    }

    padding = max(len(x) for x in tests)

    # store creation platform in cache
    found = {'platform': sys.platform.lower()}

    # detect devices on ports
    _ddop_pf = functools.partial(detect_device_on_port, tests=tests, configs=configs)
    with cf.ThreadPoolExecutor() as executor:
        detected = (executor.submit(_ddop_pf, port) for port in ports)
        for future in cf.as_completed(detected):
            key, settings, value = future.result()
            if key not in found:
                found[key] = (settings, [value])
            else:
                found[key][1].append(value)

            display_found(key, value, padding)

    # write to cache
    if not args.nocache:
        common.configcache_write(found, common.DEVICE_CACHE)


##############################################################################
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
Generates two plots from a pandas DataFrame.

The first shows raw acquired data, the second provide distributions of sampled
data.
"""

import argparse
import datetime
import itertools
import os

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tables

from mmcb import common


##############################################################################
# command line option handler
##############################################################################

def check_arguments():
    """
    handle command line options

    --------------------------------------------------------------------------
    args : none
    --------------------------------------------------------------------------
    returns
        parser.parse_args()
    --------------------------------------------------------------------------
    """
    parser = argparse.ArgumentParser(
        description='Reads in a binary data file containing a Pandas\
        DataFrame with environmental sensing data, then generates plots of\
        the raw data and summary statistics.')
    parser.add_argument(
        'filename', nargs=1, metavar='filename',
        help='Specify the file to plot. Supported file extensions:\
        h5 (HDF5 with Blosc Zstandard compression),\
        dat (Pandas CSV with ZIP compression),\
        pbz2 (Python Pickle, bzip2 compression).',
        type=str)

    return parser.parse_args()


##############################################################################
# plot
##############################################################################

def plot_raw(data, units, infile, unique_units, colour_map, marks):
    """
    Creates a raw data plot with one subplot for each unique unit of
    measurement for all sensors.

    --------------------------------------------------------------------------
    args
        data : pandas.DataFrame
        units : string
            description of x-axis (time) units
        infile : string
            filename of stored pandas dataframe
        unique_units : set
            A separate subplot is created for each unit in this set.
            e.g. {'DP°C', 'hPa', 'RH%', 'kPa', '°C'}
        colour_map : dict
            Mapping of column names to colours, where the colour for any given
            column is unique to its sensor. This ensures that sensor colours
            remain constant over multiple subplots.
            e.g.
            {'ntc °C': (0.839, 0.152, 0.156, 1.0),
             'hyt221_ch0_M1 DP°C': (0.172, 0.627, 0.172, 1.0), ...}
    --------------------------------------------------------------------------
    returns : no explicit return
            plot written to local filestore
    --------------------------------------------------------------------------
    """
    lwi = 0.5
    lal = 1

    y_axis_label_prefix = {
        '°C': 'temperature (°C)',
       'RH%': 'relative humidity (RH%)',
       'DP°C': 'calculated dew point (°C)',
       'kPa': 'vacuum (kPa)',
       'lps': 'flow rate (l/s)',
       'hPa': 'pressure (hPa)',
    }

    num_plots = len(unique_units)
    # for 3 subplots a 12 x 8 size works well, use this as a basis for other
    # numbers of plots
    y_size = (8 / 3) * num_plots

    fig, axes = plt.subplots(nrows=num_plots, ncols=1, sharex=True, figsize=(12, y_size))

    # avoid crash when there is only one subplot
    try:
        len(axes)
    except TypeError:
        axes = [axes]

    for index, unit in enumerate(unique_units):
        ccc = (c for c in data.columns if unit == c.split(' ')[-1])
        columns = sorted(ccc, key=lambda x: x.lower())
        mcmap = [
            colour_map.get(c.split(' ')[0], '#000000')
            for c in columns
        ]
        myax = data.plot(
            kind='line', x='timestamp', y=columns, ax=axes[index],
            alpha=lal, linewidth=lwi, color=mcmap
        )

        axes[index].legend(loc='center left', bbox_to_anchor=(1.01, 0.5))

        yal_prefix = y_axis_label_prefix.get(unit)
        yal = unit if yal_prefix is None else yal_prefix
        axes[index].set_ylabel(yal)

        # add day marks (if any)
        ylims = myax.get_ylim()
        for mark, day_name in marks.items():
            myax.vlines(
                mark, ylims[0], ylims[1],
                zorder=-1.0, linewidth=0.5, colors='gray', alpha=0.5
            )
            # add day names to top plot
            if index == 0:
                myax.annotate(
                    day_name,
                    xy=(mark, ylims[1]),
                    xytext=(0, 5),
                    textcoords='offset points',
                    color='gray',
                    size='xx-small'
                )
            myax.set_ylim(ylims)

    axes[-1].set_xlabel(f'elapsed time ({units})')

    fig.suptitle(infile)

    plt.tight_layout()

    basename_no_ext = os.path.splitext(os.path.basename(infile))[0]
    plt.savefig(f'{basename_no_ext}_raw.pdf')


def plot_summary(data, units, infile, unique_units):
    """
    Creates a summary statistics plot with the same basic structure as that
    created by plot_raw().

    --------------------------------------------------------------------------
    args
        data : pandas.DataFrame
        units : string
            description of x-axis (time) units
        infile : string
            filename of stored pandas dataframe
        unique_units : set
            A separate subplot is created for each unit in this set.
            e.g. {'DP°C', 'hPa', 'RH%', 'kPa', '°C'}
    --------------------------------------------------------------------------
    returns : no explicit return
            plot written to local filestore
    --------------------------------------------------------------------------
    """
    y_axis_label_prefix = {
        '°C': 'temperature (°C)',
       'RH%': 'relative humidity (RH%)',
       'DP°C': 'calculated dew point (°C)',
       'kPa': 'vacuum (kPa)',
       'lps': 'flow rate (l/s)',
       'hPa': 'pressure (hPa)',
    }

    num_plots = len(unique_units)

    # for 3 subplots a 12 x 12 size works well, use this as a basis for other
    # numbers of plots
    y_size = 4.5 * num_plots

    fig, axes = plt.subplots(nrows=num_plots, ncols=1, figsize=(12, y_size),
                             constrained_layout=True)

    # avoid crash when there is only one subplot
    try:
        len(axes)
    except TypeError:
        axes = [axes]

    for index, unit in enumerate(unique_units):
        ccc = (c for c in data.columns if unit == c.split(' ')[-1])
        columns = sorted(ccc, key=lambda x: x.lower())
        data.plot(kind='box', y=columns, ax=axes[index], showfliers=False)

        yal_prefix = y_axis_label_prefix.get(unit)
        yal = unit if yal_prefix is None else yal_prefix
        axes[index].set_ylabel(yal)

    axes[-1].set_xlabel('sensor')

    acquisition_duration = f'{data["timestamp"].max():.1f}'
    if acquisition_duration == '1.0':
        units = units[:-1]
    fig.suptitle(f'{infile}: {acquisition_duration} {units} of data')

    # descriptive x-axis labels may be long and collide unless angled.
    for i in range(num_plots):
        for tick in axes[i].get_xticklabels():
            tick.set_rotation(30)

    basename_no_ext = os.path.splitext(os.path.basename(infile))[0]
    plt.savefig(f'{basename_no_ext}_stat.pdf')


##############################################################################
# utilities
##############################################################################

def time_axis_adjustment(data, marks):
    """
    Generate the parameters necessary to transform absolute UNIX-style epoch
    timestamps to relative timestamps with human-readable units.

    --------------------------------------------------------------------------
    args
        data : pandas.DataFrame
        marks : dict
    --------------------------------------------------------------------------
    returns
        units : string
        data : pandas.DataFrame
            no explicit return, mutable type amended in place
    --------------------------------------------------------------------------
    """
    earliest_timestamp = data['timestamp'].min()
    latest_timestamp = data['timestamp'].max()

    minutes_of_available_data = (latest_timestamp - earliest_timestamp) / 60

    if minutes_of_available_data > 180:
        units = 'hours'
        scale = 3600
    else:
        units = 'minutes'
        scale = 60

    data['timestamp'] = (data['timestamp'] - earliest_timestamp) / scale

    marks_adjusted = {(k - earliest_timestamp) / scale: v for k, v in marks.items()}

    return units, marks_adjusted


##############################################################################
# main
##############################################################################

def main():
    """
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.plot.html
    """
    # set matplotlib defaults
    matplotlib.rcParams.update({
        'legend.fontsize': 'x-small',
        # fontsize of title
        'legend.title_fontsize': 'x-small'})

    args = check_arguments()
    infile = args.filename[0]
    print(f'reading {infile}')
    file_extension = os.path.splitext(infile)[-1]

    if file_extension == '.h5':
        try:
            with pd.HDFStore(infile, 'r') as hdf:
                data = pd.DataFrame(hdf['key'])
        except tables.exceptions.HDF5ExtError:
            # No need to show the user the HDF5 error backtrace.
            # The filename has alread been printed.
            print('unable to open file')
            return
    elif file_extension == '.dat':
        # deprecated file format - support retained for archived data
        data = pd.read_csv(infile, compression='zip')
    elif file_extension == '.pbz2':
        # deprecated file format - support retained for archived data
        data = common.data_read(infile)
    else:
        print(f'unrecognised file extension {file_extension}')
        return

    ##########################################################################
    # generate dew point columns for appropriate sensors

    try:
        columns_of_interest = {
            x for x in data.columns if 'RH%' in x or '°C' in x
        }
    except NameError:
        # no data was read in from the user-specified file
        return

    sensor_rhp = {c.split(' ')[0] for c in columns_of_interest if 'RH%' in c}
    sensor_dgc = {c.split(' ')[0] for c in columns_of_interest if '°C' in c}
    sensor_dew = sensor_rhp.intersection(sensor_dgc)

    # Avoid divide by zero in dew_point function later, since the BME680 seems
    # to return zero for relative humidity in dry conditions due to limited
    # resolution.
    for column in {c for c in columns_of_interest if 'RH%' in c}:
        data[column].replace(0.0, 0.001, inplace=True)

    # create new columns with calculated dew points
    for sensor in sensor_dew:
        source_columns = {x for x in columns_of_interest if sensor in x}

        try:
            column_temp = next(x for x in source_columns if '°C' in x)
            column_humi = next(x for x in source_columns if 'RH%' in x)
        except StopIteration:
            continue

        new_column_title = f'{sensor} DP°C'

        # problems when BME680 RH% = 0
        data[new_column_title] = common.dew_point(data[column_temp], data[column_humi])

    # Since RH% values at zero were changed to a small non-zero value above,
    # clamp the dew point values to -71 which is the figure for the dry air
    # supply in the cleanroom.
    for column in {x for x in data.columns if 'DP°C' in x}:
        data[column] = np.where((data[column] < -71), -71.0, data[column])

    ##########################################################################
    # work out how many plots are required - one for each different unit

    unique_units = {x.split(' ')[-1] for x in data.columns}
    unique_units.discard('timestamp')

    ##########################################################################
    # colour map - make sure sensor colours are the same in all subplots
    #
    # https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.plot.line.html
    # https://matplotlib.org/stable/tutorials/colors/colormaps.html

    # get all unique sensor names and assign a colour
    unique_sensor_name = {x.split(' ')[0] for x in data.columns if x != 'timestamp'}
    colours = [plt.cm.tab10(x) for x in range(10)]
    colour_map = dict(zip(sorted(unique_sensor_name), itertools.cycle(colours)))

    # assign each dataframe column a colour based on its sensor
    cmap = {}
    columns = {x for x in data.columns if x != 'timestamp'}
    for column in columns:
        try:
            matching_sensor = next(x for x in unique_sensor_name if x in column)
        except StopIteration:
            continue

        cmap[column.split(' ')[0]] = colour_map[matching_sensor]

    ##########################################################################
    # plot

    start_timestamp = min(data.timestamp)
    start_date = datetime.datetime.fromtimestamp(start_timestamp)
    end_timestamp = max(data.timestamp)
    end_date = datetime.datetime.fromtimestamp(end_timestamp)

    year, month, day = start_date.year, start_date.month, start_date.day
    mark = datetime.datetime(year=year, month=month, day=day) + datetime.timedelta(days=1)

    marks = {}
    while mark < end_date:
        date_ref = pd.to_datetime(mark)
        dayname = date_ref.day_name()[0]
        day = date_ref.day
        month = date_ref.month
        year = str(date_ref.year)[-2:]
        date_string = f'{year}-{month}-{day} ({dayname})'
        marks[datetime.datetime.timestamp(mark)] = date_string
        mark += datetime.timedelta(days=1)

    if len(marks) > 30:
        marks.clear()

    # Adjust the x-axis so it starts at zero and is in human-readable units.
    units, marks = time_axis_adjustment(data, marks)

    # Creates one subplot for each unique unit of measurement for all sensors.
    plot_raw(data, units, infile, unique_units, cmap, marks)

    # Summary statistics plot with the same basic structure as the plot above.
    plot_summary(data, units, infile, unique_units)


##############################################################################
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
Reads in a file containing a Pandas DataFrame (environmental sensing data)
with either a .pbz2 (compressed Python Pickle) file or a .dat extension.
Creates a ROOT file containing the raw data, processed data, and plots for
each quantity against time.
"""

import argparse
import array
import math
import os

import numpy as np
import pandas as pd
import ROOT

from mmcb import common


##############################################################################
# command line option handler
##############################################################################

def check_arguments():
    """
    handle command line options

    --------------------------------------------------------------------------
    args : none
    --------------------------------------------------------------------------
    returns
        parser.parse_args()
    --------------------------------------------------------------------------
    """
    parser = argparse.ArgumentParser(
        description='Reads in a file containing environmental sensing data.\
        Creates a ROOT file containing the raw data, processed data, and\
        plots for each quantity against time.')
    parser.add_argument(
        'filename', nargs=1, metavar='filename',
        help='Specify the file to convert. Supported file extensions:\
        h5 (HDF5 with Blosc Zstandard compression),\
        dat (Pandas CSV with ZIP compression),\
        pbz2 (Python Pickle, bzip2 compression).',
        type=str)

    return parser.parse_args()


##############################################################################
# utilities
##############################################################################

def clean(column_title):
    """
    Clean up (simplify) column titles so they'll be displayed without issue
    by rootbrowse.

    --------------------------------------------------------------------------
    args
        column_title : string
            e.g. 'TCt1_M1_flex °C'
    --------------------------------------------------------------------------
    returns
        ct : string
            e.g. 'TCt1_M1_flex_degC'
    --------------------------------------------------------------------------
    """
    ct = column_title.replace('°', 'deg').replace('%', 'PC').replace(' ', '_')
    ct = ct.encode('ascii', errors='ignore').decode("ascii")

    return ct


##############################################################################
# main
##############################################################################

def main():
    """
    Reads in a file containing a Pandas DataFrame (environmental sensing data)
    with either a .pbz2 (compressed Python Pickle) file or a .dat extension.
    Creates a ROOT file containing the raw data, processed data, and plots
    for each quantity against time.
    """

    ##########################################################################
    # read file
    ##########################################################################

    args = check_arguments()
    infile = args.filename[0]
    print(f'reading {infile}')
    file_extension = os.path.splitext(infile)[-1]

    if file_extension == '.h5':
        with pd.HDFStore(infile, 'r') as hdf:
            data = pd.DataFrame(hdf['key'])
    elif file_extension == '.dat':
        # deprecated file format - support retained for archived data
        data = pd.read_csv(infile, compression='zip')
    elif file_extension == '.pbz2':
        # deprecated file format - support retained for archived data
        data = common.data_read(infile)
    else:
        print(f'unrecognised file extension {file_extension}')
        return

    ##########################################################################
    # make sure rootls -t <filename> gives identical results whether the .dat
    # was written by this script or by sense.py
    ##########################################################################

    try:
        data = data[sorted(data.columns)]
    except NameError:
        # no data was read in from the user-specified file
        return

    ##########################################################################
    # save raw data
    ##########################################################################

    outfile = f'{os.path.splitext(infile)[0]}.root'
    print(f'writing {outfile}')
    rootfile = ROOT.TFile.Open(outfile, 'RECREATE', 'data acquired')

    # use ZSTD compression level 5 (int value 505) for the whole file
    # see: https://root.cern.ch/doc/master/Compression_8h_source.html
    rootfile.SetCompressionSettings(ROOT.RCompressionSetting.EDefaults.kUseGeneralPurpose)
    rootfile.cd()

    column_titles_text = ':'.join(clean(d) for d in data.columns)
    ntuple_raw = ROOT.TNtuple('raw_data', 'Atlas RD53 Environmental Sensing (RAW)',
                              column_titles_text)

    for _index, row_contents in data.iterrows():
        ntuple_raw.Fill(array.array('f', row_contents))

    ##########################################################################
    # generate dew point column(s)
    #
    # ... but only for sensors that generate all the required data
    # themselves. Data from different sensors are not mixed.
    ##########################################################################

    columns_of_interest = {x for x in data.columns if 'RH%' in x or '°C' in x}

    # shortlist sensors suitable for dew point calculation
    sensor_rhp = {c.split(' ')[0] for c in columns_of_interest if 'RH%' in c}
    sensor_dgc = {c.split(' ')[0] for c in columns_of_interest if '°C' in c}
    sensor_dew = sensor_rhp.intersection(sensor_dgc)

    # Avoid divide by zero in dew_point function later, since the BME680 seems
    # to frequently return zero for relative humidity in dry conditions.
    for column in {c for c in columns_of_interest if 'RH%' in c}:
        data[column].replace(0.0, 0.001, inplace=True)

    # create new columns with calculated dew points
    for sensor in sorted(sensor_dew):
        source_columns = {x for x in columns_of_interest if sensor in x}

        try:
            column_temp = next(x for x in source_columns if '°C' in x)
            column_humi = next(x for x in source_columns if 'RH%' in x)
        except StopIteration:
            continue

        new_column_title = f'{sensor} DP°C'
        data[new_column_title] = common.dew_point(data[column_temp], data[column_humi])

    # Since RH% values at zero were changed to a small non-zero value above,
    # clamp the dew point values to -71 which is the figure for the dry air
    # supply in the cleanroom.
    for column in {x for x in data.columns if 'DP°C' in x}:
        data[column] = np.where((data[column] < -71), -71.0, data[column])

    ##########################################################################
    # adjust the x-axis so it starts at zero and is in human readable units
    ##########################################################################

    units = common.time_axis_adjustment(data)

    ##########################################################################
    # store processed and calculated data
    ##########################################################################

    column_titles_text = ':'.join(clean(d) for d in data.columns)
    ntuple = ROOT.TNtuple('processed_data', 'Atlas RD53 Environmental Sensing (PROCESSED)',
                          column_titles_text)

    for _index, row_contents in data.iterrows():
        ntuple.Fill(array.array('f', row_contents))

    ##########################################################################
    # create plots from the processed and calculated data
    ##########################################################################

    ts = data['timestamp'].tolist()
    tsa = array.array('f', ts)

    for column in (c for c in data.columns if c != 'timestamp'):
        unit = column.split(' ')[-1]
        if 'DP°C' in unit:
            unit = 'calculated dew point (#circC)'
        elif '°' in unit:
            unit = 'temperature (#circC)'
        elif 'RH%' in unit:
            unit = f'humidity ({unit})'
        elif 'Pa' in unit:
            unit = f'pressure ({unit})'

        # don't ask ROOT to plot data with NaN values
        values = array.array('f', data[column].tolist())
        ts = []
        va = []
        for t, v in zip(tsa, values):
            if not math.isnan(v):
                ts.append(t)
                va.append(v)

        graph = ROOT.TGraph(len(ts), array.array('f', ts), array.array('f', va))
        graph.GetXaxis().CenterTitle()
        graph.GetYaxis().CenterTitle()
        clean_column = clean(column)
        graph.SetTitle(f'{clean_column};elapsed time ({units});{unit}')
        graph.Write(clean_column)

    # Write the data from the TFile to the actual file on disk.
    rootfile.cd()
    rootfile.Write()
    rootfile.Close()


##############################################################################
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
Environmental data logging for the ATLAS inner tracker (ITK) pixels
multi-module cycling box.

This script is expected to run on a Raspberry Pi.
"""

import argparse
import contextlib
import datetime
import functools
import gc
import logging
import os
import time
import warnings

import numpy as np
import pandas as pd
from tables import NaturalNameWarning
import zmq

from mmcb.common import UNIT
from mmcb import configure_environment as ce


##############################################################################
# command line option handler
##############################################################################


def check_file_exists(filename):
    """
    check if file exists

    --------------------------------------------------------------------------
    args
        val : string
            filename, e.g. 'config.txt'
    --------------------------------------------------------------------------
    returns : string
    --------------------------------------------------------------------------
    """
    if not os.path.exists(filename):
        raise argparse.ArgumentTypeError(f'{filename}: file does not exist')

    return filename


def check_arguments():
    """
    Handle command line options.

    --------------------------------------------------------------------------
    args : none
    --------------------------------------------------------------------------
    returns : class argparse.ArgumentParser
    --------------------------------------------------------------------------
    """
    parser = argparse.ArgumentParser(
        description='ATLAS inner tracker (ITK) pixels multi-module cycling\
        box environmental logging script. This script logs various parameters\
        including temperature, humidity and pressure from various sensors.\
        Dew points are calculated for all sensors that supply both humidity\
        and temperature; these calculated values are echoed to the screen for\
        information, but are not written to the log file. Sensors are\
        connected by I2C in the current test setup, mostly via Qwiic I2C\
        multiplexer(s). This script will run continuously until terminated.\
        The human-readable log file on mass storage will be updated every\
        data point, and raw data is also streamed to an HDF5 file. If this\
        script was started from a remote client with nohup then use:\
        kill -s INT <PID> or kill -SIGINT <PID> to terminate the script.'
    )
    parser.add_argument(
        'config_filename',
        nargs=1,
        metavar='configuration file',
        help='Specify the file containing the test setup configuration.',
        type=check_file_exists,
        default=None,
    )
    parser.add_argument(
        '--label',
        nargs=1,
        metavar='filename_label',
        help='Add a label suffix to the timestamp when writing to filenames.\
        Use speech marks around the label if it contains spaces.',
        default=None,
    )

    return parser.parse_args()


##############################################################################
# utilities
##############################################################################


def environment_string(padding, sensor_details, sensor_reading):
    """
    Create a string for the given sensor and reading such that values will
    align in sequential log entries for the given unit.

    --------------------------------------------------------------------------
    args
        padding : int
            make some effort to keep entries in sequential log lines aligned
        sensor_details : string
            e.g. 'hyt_M1 °C'
        sensor_reading : float
            e.g. 20.438564365500824
    --------------------------------------------------------------------------
    returns : string
        e.g.
            'hyt_M1   20.438 °C'
    --------------------------------------------------------------------------
    """
    name, _, unit = sensor_details.rpartition(' ')
    return f'{name} {sensor_reading:{padding}.3f} {unit}'


def fill_missing(line, reference_fields):
    """
    Make sure the line dict has key/value pairs for every element(key) in set
    reference_fields.

    --------------------------------------------------------------------------
    args
        line : dict
            holds all data points from all connected sensors for the current
            timestamp
        reference_fields : set
            e.g. {'AMBIENT °C', 'OVEN_AIR °C', 'timestamp'}
    --------------------------------------------------------------------------
    returns : none
        line : dict
            mutable type amended in place, no explicit return
    --------------------------------------------------------------------------
    """
    for field in reference_fields:
        if field not in line:
            line[field] = np.nan


def timestamp_to_utc(tref):
    """
    Converts a timestamp into a string in UTC to the nearest second.

    e.g. 1567065212.1064236 converts to '20190829_075332'

    --------------------------------------------------------------------------
    args
        tref : float
            time in seconds since the epoch
    --------------------------------------------------------------------------
    returns : string
    --------------------------------------------------------------------------
    """
    utc = datetime.datetime.utcfromtimestamp(tref).isoformat().split('.')[0]
    return utc.replace('-', '').replace(':', '').replace('T', '_')


##############################################################################
# generate log lines from dict
##############################################################################


def generate_log_lines(enviro_string, line):
    """
    Generate all the necessary log lines for one point in time. Temperature,
    humidity and pressure each have their own log lines.

    --------------------------------------------------------------------------
    args
        enviro_string : functools.partial
            function that takes a data point and returns a formatted string
            suitable for use as part of a log line
        line : dict
            holds all data points from all connected sensors for the current
            timestamp
    --------------------------------------------------------------------------
    returns : none
        no explicit return: lines written to log file
    --------------------------------------------------------------------------
    """
    for unit in UNIT.__dict__.values():
        string = ', '.join(
            enviro_string(sensor, reading)
            for sensor, reading in line.items()
            if unit in sensor
        )
        if string:
            if UNIT.calculated_dew_point in string:
                # show on-screen only, do not write to log file
                logging.debug(string)
            else:
                logging.info(string)


##############################################################################
# share data with iv.py
##############################################################################


def emit_data(message_tx, zmq_skt):
    """
    Receive acquired environmental data and send it to iv.py.

    To avoid dropped data packets ZeroMQ REQ/REP is the pattern used, though
    the desired behaviour is really PUB/SUB. PUB/SUB is unreliable in its
    basic form, and can't be used. With REQ/REP iv.py is obliged to
    reply to the message this thread sends, but that reply can be discarded.

    --------------------------------------------------------------------------
    args
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        zmq_skt : zmq.sugar.socket.Socket
            ZeroMQ socket to communicate with iv.py
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    # send message then attempt to receive reply
    try:
        zmq_skt.send_json(message_tx)
    except zmq.error.ZMQError:
        pass
    else:
        # receive and discard reply from iv.py
        with contextlib.suppress(zmq.error.ZMQError):
            zmq_skt.recv_json()


##############################################################################
def main():
    """
    Read the test setup from a configuration file.
    Build lists of sensors and initialise everything from there.
    Run.
    """

    # This suppresses warnings issued due to characters such as degree
    # symbols being used in column titles, e.g. 'OVEN_AIR °C'. Certain access
    # methods within PyTables rely on these strings matching this regex
    # ^[a-zA-Z_][a-zA-Z0-9_]*$ though non-matching column titles seems to be
    # tolerated reading/writing hdf5 files via Pandas.
    warnings.filterwarnings('ignore', category=NaturalNameWarning)

    ##########################################################################
    # read command line arguments
    ##########################################################################

    args = check_arguments()

    ##########################################################################
    # enable logging to file and screen
    ##########################################################################

    # date and time to the nearest second when the script was started
    filename = timestamp_to_utc(time.time())

    if args.label is not None:
        filename = f'{filename}_{args.label[0].replace(" ", "_")}'

    formatter = logging.Formatter('%(asctime)s : %(levelname)s : %(message)s')
    # set the minimum log-level required by any subsequent handler
    logging.getLogger('').setLevel(logging.DEBUG)

    # log to file
    fha = logging.FileHandler(f'{filename}.log')
    fha.setLevel(logging.INFO)
    fha.setFormatter(formatter)
    logging.getLogger('').addHandler(fha)

    # log to screen
    # all calculated figures are DEBUG and therefore shown on screen only
    cha = logging.StreamHandler()
    cha.setLevel(logging.DEBUG)
    cha.setFormatter(formatter)
    logging.getLogger('').addHandler(cha)

    # set logging time to UTC to match session timestamp
    logging.Formatter.converter = time.gmtime

    ##########################################################################
    # zeromq simple request-reply to send environmental data to iv.py
    ##########################################################################

    context = zmq.Context()
    zmq_skt = context.socket(zmq.REQ)
    # zmq.RCVTIMEO and zmq.SNDTIMEO specified in units of milliseconds
    zmq_skt.setsockopt(zmq.RCVTIMEO, 200)
    zmq_skt.setsockopt(zmq.SNDTIMEO, 200)
    zmq_skt.setsockopt(zmq.DELAY_ATTACH_ON_CONNECT, 1)
    # iv.py uses 5555 for communication with liveplot.py
    with contextlib.suppress(zmq.error.ZMQError):
        zmq_skt.connect("tcp://localhost:5556")

    send_data_to_iv_script = functools.partial(emit_data, zmq_skt=zmq_skt)

    ##########################################################################
    # set up environment
    ##########################################################################

    print('Configure test environment: start')
    testenv = ce.TestEnvironment(args.config_filename[0])
    print('Configure test environment: complete')

    padding = len('1000.000')
    enviro_string = functools.partial(environment_string, padding)
    dict_to_log_lines = functools.partial(generate_log_lines, enviro_string)

    ##########################################################################
    # define reference fields
    #
    # The reference fields will be all the keys returned by
    # read_all_sensors(), where key/value pairs represent the sampled
    # parameters and their values for any given timestamp.
    #
    # example returned values from read_all_sensors():
    # {
    #     'timestamp': 1650734619.065016,
    #     'smc_vacuum': -0.3395540799999992,
    #     'ntc_temperature': -61.19908797184263,
    #     ...
    # }
    #
    # for which fields will be {'timestamp', 'smc_vacuum', 'ntc_temperature'}
    #
    ##########################################################################

    print('Initial sensor read')

    fields = set(testenv.read_all_sensors())

    if not fields:
        print('No sensors found, exiting.')
        return

    ##########################################################################
    # acquire data
    ##########################################################################

    print('Acquiring data')
    with pd.HDFStore(f'{filename}.h5', mode='w', complib='blosc:zstd', complevel=5) as datafile:
        with contextlib.suppress(KeyboardInterrupt):
            while True:
                time_0 = time.monotonic()

                measurements = testenv.read_all_sensors()

                send_data_to_iv_script(measurements)
                dew_points = testenv.calculate_dew_points(measurements)
                measurements.update(dew_points)
                dict_to_log_lines(measurements)

                # fields must be identical for when writing to hdf5 files so add
                # any missing fields
                fill_missing(measurements, fields)

                datafile.append(
                    'key',
                    pd.DataFrame([measurements]),
                    format='table',
                    data_columns=True
                )

                # Force garbage collection. Without this tracemalloc indicates
                # ~/pve/lib/python3.9/site-packages/tables/filters.py:184
                # adds ~3KiB per loop, which can be troublesome for data
                # acquisitions that run for a few weeks.
                gc.collect()

                tdif = time.monotonic() - time_0
                if tdif < 10:
                    time.sleep(10 - tdif)


##############################################################################
if __name__ == '__main__':
    main()



"""
Common functions related to serial port commands for project devices.

The functions motion() and power() are relatively expensive function calls
as there is a fair amount of substitution to perform for each call. Calls for
write commands (with variable parameters) will cache poorly, e.g. setting
voltage values; constant commands that read voltage/current will cache well.

power() has an LRU cache applied as calls are made very frequently to read
power supply voltage/current/status. Calls to motion(), though similarly
expensive, are relatively sparse.

e.g. for command:

import statistics
import timeit
statistics.mean(timeit.Timer("lexicon.power('2410', 'read measured vi')",
                             setup='import lexicon').repeat(100,1))

timings are:

    2.043e-05s (un-cached)
    4.207e-07s (cached)

Cache performance for: ./iv.py -200 --hold 1

CacheInfo(hits=82, misses=56, maxsize=128, currsize=56)

callable functions from this file:

    motion
    power
"""

import functools


##############################################################################
# motion controller command string generation
##############################################################################

def motion(controller, command, identifier=None, argument1=None, argument2=None):
    """
    generate the requested RS232 command sequence appropriate for the
    motion controller specified

    --------------------------------------------------------------------------
    args
        controller : string
            name of the controller
        command : string
            name of the command to be fetched
        identifier : int
            this may be either the controller axis to which the stage is
            connected (for esp30 or mm4006), or the controller address in the
            case of the smc100
        argument1 : numeric
            a parameter to substitute into the command string
        argument2 : numeric
            a parameter to substitute into the command string
    --------------------------------------------------------------------------
    returns : string
        a valid command that can be sent to the controller over RS232
    --------------------------------------------------------------------------
    """
    terminate = '\r\n'
    axis = address = identifier

    ##########################################################################
    # Language: proprietary
    #
    # "The maximum number of characters allowed on a command line is 80."
    # ESP300.pdf, p. 3-8
    esp30 = {
        'is stage moving': f'{axis}MD?',
        'motor on': f'{axis}MO',
        # p.3.98, response: 1 (ON) 0 (OFF)
        'check motor on': f'{axis}MO?',
        # argument1: decimal.Decimal, absolute position
        'move absolute': f'{axis}PA{argument1}',
        # argument1: decimal.Decimal, relative distance to move
        'move relative': f'{axis}PR{argument1}',
        'move to home position': f'{axis}MO;{axis}OR1',
        'read controller version': 'VE',
        'read error': 'TE?',
        'read position': f'{axis}TP?',
        'read stage identifier': f'{axis}ID?',
        'set speed': (
            # argument1: int, acceleration
            # argument2: int, velocity
            f'{axis}AC{argument1};'
            f'{axis}AG{argument1};'
            f'{axis}VU{argument2};'
            f'{axis}VA{argument2};'
            f'{axis}OL{argument2};'
            f'{axis}OH{argument2}{terminate}'
            f'{axis}AC?;'
            f'{axis}AG?;'
            f'{axis}VU?;'
            f'{axis}VA?;'
            f'{axis}OL?;'
            f'{axis}OH?'),
        'software limits check': f'{axis}ZS?',
        'software limits disable': f'{axis}ZS4H',
        'stop motion': f'{axis}ST'}

    ##########################################################################
    # Language: proprietary
    #
    # "A single command line ... may not exceed 110 characters."
    # MM4K6_701_E3.pdf, p. 3.16
    mm4006 = {
        'is stage moving': f'{axis}TS',
        'motor on': f'{axis}MO',
        # check motor status, MM4K6_701_E3.pdf, p.3.86
        # bit 1 motor status: 0 (ON) 1 (OFF)
        'check motor on': f'{axis}MS',
        # argument1: decimal.Decimal, absolute position
        'move absolute': f'{axis}PA{argument1}',
        # argument1: decimal.Decimal, relative distance to move
        'move relative': f'{axis}PR{argument1}',
        'move to home position': f'{axis}MO;{axis}OR',
        'read controller version': 'VE',
        'read error': 'TE',
        'read position': f'{axis}TP',
        'read stage identifier': f'{axis}TS',
        'set speed': (
            # argument1: int, acceleration
            # argument2: int, velocity
            f'{axis}AC{argument1};'
            f'{axis}VU{argument2};'
            f'{axis}VA{argument2};'
            f'{axis}OL{argument2};'
            f'{axis}OH{argument2}{terminate}'
            f'{axis}AC?;'
            f'{axis}VU?;'
            f'{axis}VA?;'
            f'{axis}OL?;'
            f'{axis}OH?'),
        'stop motion': f'{axis}ST'}

    ##########################################################################
    # Language: proprietary
    smc = {
        'is stage moving': f'{address}TS',
        'motor on': f'{address}MM1',
        'check motor on': f'{address}MM1?',
        # argument1: decimal.Decimal, absolute position
        'move absolute': f'{address}PA{argument1}',
        # argument1: decimal.Decimal, relative distance to move
        'move relative': f'{address}PR{argument1}',
        'move to home position': f'{address}TS{terminate}{address}OR',
        'move to home position 2': '',
        'read controller version': f'{address}VE',
        'read error': f'{address}TE',
        'read position': f'{address}TP',
        'read stage identifier': f'{address}ID?',
        'stop motion': f'{address}ST'}

    instructions = {'esp300': esp30, 'mm4006': mm4006, 'smc': smc}

    # sanity checks
    first = instructions.get(controller)
    assert first is not None, 'unrecognised controller name'
    second = first.get(command)
    assert second is not None, 'unrecognised command'
    cmd_string = instructions[controller][command]
    assert 'None' not in cmd_string, 'missing argument to function'

    return bytes(cmd_string + terminate, 'utf-8')


##############################################################################
# power supply command string generation
##############################################################################

@functools.lru_cache(maxsize=64)
def power(model, command, argument1=None, argument2=None, channel=None):
    """
    generate the requested RS232 command sequence appropriate for the
    power supply specified

    Manufacturers' implementations of Standard Commands for Programmable
    Instruments (SCPI) all seem to have their own quirks, which is why there
    is no generic SCPI command sequence, and each power supply has its own
    dictionary.

    --------------------------------------------------------------------------
    args
        model : string
            power supply model
        command : string
            name of the command to be fetched
        argument1 : numeric/string
            a parameter to substitute into the command string
        argument2 : numeric/string
            a parameter to substitute into the command string
        channel : string
            channel name, the caller is required to use this for
            multichannel power supplies, but may omit it for single-channel
    --------------------------------------------------------------------------
    returns : string
        a valid command that can be sent to the psu over RS232
    --------------------------------------------------------------------------
    """
    terminate = '\n' if model == '2614b' else '\r\n'

    ##########################################################################
    # Language: Standard Commands for Programmable Instruments (SCPI)
    keithley_2410 = {
        # returns '0' (not in compliance) or '1' (in compliance) (p. 18-70)
        'check compliance': ':CURR:PROT:TRIP?',
        # returns '0' (disabled - no restrictions) or '1' (enabled)
        'check interlock': ':OUTP:INT:STAT?',
        # returns '0' (output off) or '1' (output on)
        'check output': ':OUTP?',
        'clear event registers': '*CLS',
        # compliance definitions on p. 6-5
        'configure': (
            # clear all event registers and error queue
            f'*CLS;'
            # when reading status registers, use decimal format
            f':FORM:SREG ASC;'
            # select voltage source
            f':SOUR:FUNC VOLT;'
            # set fixed voltage source mode
            f':SOUR:VOLT:MODE FIXED;'
            # set voltage range to 1000V
            f':SOUR:VOLT:RANG 1000;'
            # set psu to output given voltage
            f':SOUR:VOLT:LEV {argument1};'
            # set power supply's current limit
            f':CURR:PROT {argument2};'
            # set current measurement range to match compliance value
            f':CURR:RANG:UPP {argument2};'
            # and make sure the measurement range stays constant
            f':CURR:RANG:AUTO OFF'),
        'configure range': (
            # clear all event registers and error queue
            '*CLS;'
            # when reading status registers, use decimal format
            ':FORM:SREG ASC;'
            # select voltage source
            ':SOUR:FUNC VOLT;'
            # set fixed voltage source mode
            ':SOUR:VOLT:MODE FIXED;'
            # set voltage range to 1000V
            ':SOUR:VOLT:RANG 1000'),
        'identify': '*CLS;*IDN?',
        'output off': ':OUTP OFF',
        'output on': ':OUTP ON',
        'read measured vi': (
            # set concurrent measurement mode on
            f':FUNC:CONC 1;'
            # measure actual dc voltage and current
            f':FUNC:ON "VOLT:DC","CURR:DC"{terminate}'
            # read value string
            f':READ?'),
        'read set voltage': (
            # disable concurrent measurement mode
            f':FUNC:CONC 0;'
            # measure actual dc current (forces set voltage to appear in index 0)
            f':FUNC:ON "CURR:DC"{terminate}'
            # read value string
            f':READ?'),
        'read voltage': '*CLS;:READ?',
        'reset': '*RST',
        # see p. 18-69 (p. 449 in PDF)
        'get current limit': ':CURR:PROT:LEV?',
        # compliance definitions on p. 6-5
        'set current limit': (
            # set power supply's current limit
            f':CURR:PROT {argument1};'
            # set current measurement range to match compliance value
            f':CURR:RANG:UPP {argument1};'),
        # argument1: power supply output 'ON' or 'OFF'
        'set output': f':OUTP {argument1}',
        # argument1: use power supply 'REAR' or 'FRON' output
        'set route': f':ROUT:TERM {argument1}',
        'set voltage': (
            # clear all event registers and error queue
            f'*CLS;'
            # set psu voltage
            # argument1: decimal.Decimal, voltage
            f':SOUR:VOLT:LEV {argument1}'),
        'set voltage and range': (f':SOUR:FUNC VOLT;'
                                  f':SOUR:VOLT:MODE FIXED;'
                                  f':SOUR:VOLT:RANG 1000;'
                                  # argument1: decimal.Decimal, voltage
                                  f':SOUR:VOLT:LEV {argument1}')}

    ##########################################################################
    # Language: Lua
    keithley_2614b = {
        # example response: 'true' or 'false' (p. 11-238)
        'check compliance': f'print(smu{channel}.source.compliance)',
        # Without hardware interlock: '0.00000e+00'
        # with hardware interlock: '2.04800e+03'
        'check interlock': 'print(status.measurement.condition)',
        # returns '0.00000e+00' (off) or 1.00000e+00' (on)
        'check output': f'print(smu{channel}.source.output)',
        # 2614b also has separate commands for model and serial
        # e.g. m=localnode.model;s=localnode.serialno;print(m,s)
        'clear event registers': 'errorqueue.clear()',
        # argument1: decimal.Decimal, voltage
        'configure': (f'beeper.enable=beeper.OFF;'
                      f'smu{channel}.source.func=smu{channel}.OUTPUT_DCVOLTS;'
                      f'smu{channel}.source.rangev=200;'
                      f'smu{channel}.source.autorangev=smu{channel}.AUTORANGE_OFF;'
                      f'smu{channel}.source.limiti={argument2};'
                      f'smu{channel}.source.levelv={argument1};'
                      f'smu{channel}.measure.autorangei=smu{channel}.AUTORANGE_ON;'
                      f'smu{channel}.source.output=smu{channel}.OUTPUT_ON;'
                      f'display.smu{channel}.measure.func=display.MEASURE_DCAMPS'),
        'configure range': (f'beeper.enable=beeper.OFF;'
                            f'smu{channel}.source.func=smu{channel}.OUTPUT_DCVOLTS;'
                            f'smu{channel}.source.rangev=200;'
                            f'smu{channel}.source.autorangev=smu{channel}.AUTORANGE_OFF;'
                            f'smu{channel}.measure.autorangei=smu{channel}.AUTORANGE_ON;'
                            f'display.smu{channel}.measure.func=display.MEASURE_DCAMPS'),
        # example response '1.00000e-01'
        'get current limit': f'print(smu{channel}.source.limiti)',
        # argument1: decimal.Decimal, voltage
        'set current limit': f'smu{channel}.source.limiti={argument1}',
        # example response: 'Keithley Instruments SMU 2614B - 4428182'
        'identify': ('errorqueue.clear();'
                     'print(localnode.description)'),
        # argument1: string, channel 'a' or 'b'
        'output off': f'smu{channel}.source.output=smu{channel}.OUTPUT_OFF',
        # argument1: string, channel 'a' or 'b'
        'output on': f'smu{channel}.source.output=smu{channel}.OUTPUT_ON',
        'read measured vi': f'print(smu{channel}.measure.v(),'
                            f'smu{channel}.measure.i())',
        # argument1: string, channel 'a' or 'b'
        # example response: '6.43730e-13'
        'read current': f'print(smu{channel}.measure.i())',
        # example response: '4.00000e+00'
        'read set voltage': f'print(smu{channel}.source.levelv)',
        # argument1: string, channel 'a' or 'b'
        # example response: '1.00011e+01'
        'read voltage': f'print(smu{channel}.measure.v())',
        # reset channel defaults
        'reset': 'localnode.reset()',
        # argument1: string, channel 'a' or 'b'
        # argument1: decimal.Decimal, voltage value
        'set voltage': f'smu{channel}.source.levelv={argument1}',
        # used for clearing anything in the PSU's command buffer
        'terminator only': ''}

    ##########################################################################
    # Language: proprietary
    #
    # Cannot chain multiple commands for ISEG SHQ in the same command
    # string as can be done for Keithley - particularly this affects
    # initial configuration - all commands must be called individually.
    iseg_shq = {
        'check compliance': f'S{channel}',
        'check interlock': f'S{channel}',
        'check module status': f'T{channel}',
        'check output': f'S{channel}',
        'configure': '',
        # e.g. returns '484216;3.09;4000V;3mA' for ISEG SHQ 224M (2x 4kV / 2mA)
        'identify': '#',
        'read current': f'I{channel}',
        # read maximum voltage rate of change
        'read max rate of change': f'V{channel}',
        'read voltage': f'U{channel}',
        # for 100V ('D1=100') returns 01000-01, for 1100V (D1='1100') 11000-01
        'read set voltage': f'D{channel}',
        # part of 'configure' for ISEG SHQ
        # set auto ramp
        # (as soon as voltage is set with D command, move to that voltage)
        'set auto ramp': f'A{channel}=8',
        # set inter-character delay to minimum (1ms)
        'set char delay': 'W=1',
        # Unlike Keithley, ISEG sets output channel polarity with hardware
        # screws on the rear panel.
        # All voltages specified must be positive, with a maximum of 2
        # decimal places.
        # argument1: string, channel '1' or '2'
        # argument1: decimal.Decimal, voltage value
        'set voltage': f'D{channel}={argument1}',
        # set maximum voltage rate of change to the fastest available 255V/s
        'set voltage max rate of change': f'V{channel}={argument1}',
        # ISEG SHQ requires an initial CRLF to synchronise
        'synchronise': ''}

    ##########################################################################
    # Language: Standard Commands for Programmable Instruments (SCPI)
    #
    # dual-output DC power supply 0-35V, 0.8A and 0-60V, 0.5A
    # e3647a can accept as many queries as needed with only a single read
    agilent_e3647a = {
        'check output': ':OUTP?',
        'clear event registers': '*CLS',
        'identify': '*IDN?',
        'output off': ':OUTP OFF',
        'output on': ':OUTP ON',
        'read current': f':INST:SEL OUT{channel};:MEAS:CURR?',
        'read measured vi': f':INST:SEL OUT{channel};:MEAS:VOLT?;:MEAS:CURR?',
        'read set voltage': ':VOLT?',
        'read voltage': f':INST:SEL OUT{channel};:MEAS:VOLT?',
        # argument1: decimal.Decimal, voltage value
        'set voltage': (f':INST:SEL OUT{channel};'
                        f':VOLT:RANG LOW;'
                        f':SOUR:VOLT {argument1}')}

    ##########################################################################
    # Language: Standard Commands for Programmable Instruments (SCPI)
    #
    # single-output DC power supply
    # e3634a expects response to be read back after each query and MUST be put
    # into remote mode before MEAS can be used
    agilent_e3634a = {
        'check output': ':OUTP?',
        'clear event registers': '*CLS',
        'identify': '*IDN?',
        'output off': 'OUTP OFF',
        'output on': 'OUTP ON',
        'read current': f'SYST:REM{terminate}MEAS:CURR?',
        'read set voltage': 'VOLT?',
        'read voltage': f':SYST:REM{terminate}:MEAS:VOLT?',
        'set remote': 'SYST:REM',
        # argument1: decimal.Decimal, voltage value
        'set voltage': (f'VOLT:RANG LOW{terminate}'
                        f'SOUR:VOLT {argument1}')}

    ##########################################################################
    # Language: Standard Commands for Programmable Instruments (SCPI)
    #
    # These are branded as Rohde & Schwarz but identify as Hameg
    # quad-output DC power supply, 0-32V, 10A
    hameg_hmp4040 = {
        'check output': f'INST OUT{channel}{terminate}OUTP?',
        'clear event registers': '*CLS',
        'identify': '*IDN?',
        'output off': f'INST OUT{channel}{terminate}OUTP OFF',
        'output on': f'INST OUT{channel}{terminate}OUTP ON',
        'output off all': 'OUTP:GEN OFF',
        'output on all': 'OUTP:GEN ON',
        'read current': f'INST OUT{channel}{terminate}MEAS:CURR?',
        'read set voltage': f'INST OUT{channel}{terminate}VOLT?',
        'read voltage': f'INST OUT{channel}{terminate}MEAS:VOLT?',
        'reset': '*RST',
        'set remote': 'SYST:REM',
        # argument1: decimal.Decimal, voltage value
        # note that a negative sign supplied as part of the voltage value is
        # discarded by the power supply, e.g. -4.1V will set a value of 4.1V
        'set voltage': f'INST OUT{channel}{terminate}SOUR:VOLT {argument1}',
        #
        # Additions for HMP4040 constant current mode.
        #
        # "The R&S HMP switches to the CC mode automatically, when the output
        # current exceeds the maximum value Imax"
        #     HMPSeries_UserManual_en_02.pdf, p.31 (PDF p.46)
        #
        # channel button illuminated with colour:
        #   Green : constant voltage mode (CV)
        #     Red : constant current mode (CC)
        #
        # STAT:QUES:COND? returns a 16-bit register value
        # Bit 0 is set while the instrument is in constant current mode.
        # Bit 1 is set while the instrument is in constant voltage mode.
        'read channel mode': f'STAT:QUES:INST:ISUM{channel}:COND?',
        'get current limit': f'INST OUT{channel}{terminate}CURR?',
        # argument1: float, amps (no scientific representation)
        'set current limit': f'INST OUT{channel}{terminate}CURR {argument1}',
        # argument1: float, amps
        'set current step': f'INST OUT{channel}{terminate}CURR:STEP {argument1}',
        'read current step': f'INST OUT{channel}{terminate}CURR:STEP?',
        'current step down': f'INST OUT{channel}{terminate}CURR DOWN',
        'current step up': f'INST OUT{channel}{terminate}CURR UP'}

    instructions = {
        '2410': keithley_2410, '2614b': keithley_2614b,
        'shq': iseg_shq,
        'e3647a': agilent_e3647a, 'e3634a': agilent_e3634a,
        'hmp4040': hameg_hmp4040
    }

    # sanity checks
    first = instructions.get(model)
    assert first is not None, 'unrecognised model name'
    second = first.get(command)
    assert second is not None, 'unrecognised command'
    cmd_string = instructions[model][command]
    assert 'None' not in cmd_string, 'missing argument to function'

    return bytes(cmd_string + terminate, 'utf-8')


##############################################################################
# general instrumentation
##############################################################################

@functools.lru_cache(maxsize=64)
def instrument(model, command, argument1=None, argument2=None):
    """
    generate the requested RS232 command sequence appropriate for the
    power supply specified

    Manufacturers' implementations of Standard Commands for Programmable
    Instruments (SCPI) all seem to have their own quirks, which is why there
    is no generic SCPI command sequence, and each power supply has its own
    dictionary.

    --------------------------------------------------------------------------
    args
        model : string
            instrument model
        command : string
            name of the command to be fetched
        argument1 : numeric/string
            a parameter to substitute into the command string
        argument2 : numeric/string
            a parameter to substitute into the command string
    --------------------------------------------------------------------------
    returns : string
        a valid command that can be sent to the psu over RS232
    --------------------------------------------------------------------------
    """
    terminate = '\r\n'

    ##########################################################################
    # Language: Standard Commands for Programmable Instruments (SCPI)
    #
    # Note that front/rear terminal usage can only be set using the front
    # panel button
    dmm6500 = {
        'configure to read capacitance': (
            'dmm.measure.func=dmm.FUNC_CAPACITANCE;'
            'dmm.measure.autorange=dmm.ON'
        ),
        'configure to read ac current': (
            'dmm.measure.func=dmm.FUNC_AC_CURRENT;'
            'dmm.measure.autorange=dmm.ON'
        ),
        'configure to read dc current': (
            'dmm.measure.func=dmm.FUNC_DC_CURRENT;'
            'dmm.measure.autorange=dmm.ON'
        ),
        'configure to read resistance': (
            'dmm.measure.func=dmm.FUNC_RESISTANCE;'
            'dmm.measure.autorange=dmm.ON'
        ),
        # Auto range not available for this function.
        # This is just a placeholder, and will need proper configuration later.
        # See reference manual p.14-68 (p.786 in PDF).
        'configure to read temperature': (
            'dmm.measure.func=dmm.FUNC_TEMPERATURE;'
            'dmm.measure.unit=dmm.UNIT_CELSIUS'
        ),
        'configure to read ac voltage': (
            'dmm.measure.func=dmm.FUNC_AC_VOLTAGE;'
            'dmm.measure.unit=dmm.UNIT_VOLT;'
            'dmm.measure.autorange=dmm.ON'
        ),
        'configure to read dc voltage': (
            'dmm.measure.func=dmm.FUNC_DC_VOLTAGE;'
            'dmm.measure.unit=dmm.UNIT_VOLT;'
            'dmm.measure.autorange=dmm.ON'
        ),
        'identify': '*IDN?',
        'read value': 'print(dmm.measure.read())',
        'reset': 'dmm.reset()',
    }

    instructions = {
        'dmm6500': dmm6500,
    }

    # sanity checks
    first = instructions.get(model)
    assert first is not None, 'unrecognised model name'
    second = first.get(command)
    assert second is not None, 'unrecognised command'
    cmd_string = instructions[model][command]
    assert 'None' not in cmd_string, 'missing argument to function'

    return bytes(cmd_string + terminate, 'utf-8')



#!/usr/bin/env python3
"""
Instantaneous snapshot of all power supply channels from devices connected
by FTDI USB to RS232 adaptors.

All power supplies supported by detect.py are usable by this script.
"""

import argparse
import collections
import contextlib
import sys
import time
import threading

from mmcb import common
from mmcb import lexicon


##############################################################################
# command line option handler
##############################################################################

def check_arguments(settings):
    """
    handle command line options

    --------------------------------------------------------------------------
    args : none
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    parser = argparse.ArgumentParser(
        description='Provides an instantaneous snapshot of the\
        configuration and status of power supply channels from all Keithley\
        2410, 2614b; ISEG SHQ 222M, 224M; Agilent e3647a, e3634a; and Hameg\
        (Rohde & Schwarz) HMP4040 power supplies connected by FTDI USB to\
        RS232 adaptors.')
    parser.add_argument(
        '--manufacturer', nargs=1, metavar='manufacturer',
        choices=['agilent', 'hameg', 'iseg', 'keithley'],
        help='PSU manufacturer.',
        default=None)
    parser.add_argument(
        '--model', nargs=1, metavar='model',
        choices=['2410', '2614b', 'e3634a', 'e3647a', 'hmp4040', 'shq'],
        help='PSU model.',
        default=None)
    parser.add_argument(
        '--serial', nargs=1, metavar='serial',
        help='PSU serial number. A part of the serial may be supplied if it\
        is unique amongst connected devices.',
        default=None)
    parser.add_argument(
        '--channel', nargs=1, metavar='channel',
        choices=['1', '2', '3', '4', 'a', 'A', 'b', 'B', 'c', 'C', 'd', 'D'],
        help='PSU channel number. These can be specified numerically or\
        alphabetically.',
        default=None)
    parser.add_argument(
        '--port', nargs=1, metavar='port',
        help='Serial port identifier. This is useful where multiple Agilent/\
        Keysight/Hewlett-Packard power supplies are connected, since they do\
        not provide a serial number over RS232. A part of the serial may be\
        supplied if it is unique amongst connected devices.',
        default=None)

    group1 = parser.add_mutually_exclusive_group()
    group1.add_argument(
        '--sv', action='store_true',
        help='NOT IMPLEMENTED YET. Read set voltage.\
        Only valid if the user has specified a PSU channel.',
    )
    group1.add_argument(
        '--mv', action='store_true',
        help='Read measured voltage. Only valid if the user has specified a PSU channel.',
    )
    group1.add_argument(
        '--si', action='store_true',
        help='NOT IMPLEMENTED YET. Read set current.\
        Only valid if the user has specified a PSU channel.',
    )
    group1.add_argument(
        '--mi', action='store_true',
        help='Read measured current. Only valid if the user has specified a PSU channel.',
    )

    args = parser.parse_args()

    if args.serial:
        settings['serial'] = args.serial[0]

    if args.channel:
        settings['channel'] = args.channel[0].lower()

    if args.manufacturer:
        settings['manufacturer'] = args.manufacturer[0]

    if args.model:
        settings['model'] = args.model[0]

    if args.port:
        settings['port'] = args.port[0]

    select = [
        settings['serial'],
        settings['channel'],
        settings['manufacturer'],
        settings['model'],
        settings['port'],
    ]
    settings['filter'] = any(value is not None for value in select)

    requests = collections.Counter([args.sv, args.mv, args.si, args.mi])

    if settings['filter'] and requests[True] == 0:
        print('select parameter to report with --sv, --mv, --si, --mi')
        sys.exit()

    settings['sv'] = args.sv
    settings['mv'] = args.mv
    settings['si'] = args.si
    settings['mi'] = args.mi


##############################################################################
# utilities
##############################################################################

def hmp4040_constant_voltage_current_status(ser, pipeline, dev, tcolours):
    """
    Query HMP4040 channel for constant voltage/current configuration.

    --------------------------------------------------------------------------
    args
        ser : serial.Serial
            reference for serial port
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        dev : instance of class Channel
            contains details of a device and its serial port
        tcolours : class
            contains ANSI colour escape sequences
    --------------------------------------------------------------------------
    returns : string
    --------------------------------------------------------------------------
    """
    assert dev.manufacturer == 'hameg', 'function only callable for Hameg PSU'

    command_string = lexicon.power(dev.model, 'read channel mode', channel=dev.channel)
    register = common.atomic_send_command_read_response(pipeline, ser, dev, command_string)

    status_categories = {
        # constant_voltage, constant_current
        (False, False): 'neither constant voltage nor constant current',
        (False, True): f'{tcolours.BOLD}{tcolours.FG_BRIGHT_RED}constant current{tcolours.ENDC}',
        (True, False): f'{tcolours.BOLD}{tcolours.FG_BRIGHT_GREEN}constant voltage{tcolours.ENDC}',
        (True, True): 'both constant voltage and constant current (!)',
        None: 'could not be determined'}

    configuration = None

    try:
        regval = int(register)
    except ValueError:
        pass
    else:
        constant_curr = (regval & (1 << 0)) != 0
        constant_volt = (regval & (1 << 1)) != 0
        configuration = (constant_volt, constant_curr)

    return status_categories[configuration]


def simple_current_limit(ser, pipeline, dev):
    """
    Query current limit information for HMP4040, 2410, 2614b.

    --------------------------------------------------------------------------
    args
        ser : serial.Serial
            reference for serial port
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        dev : instance of class Channel
            contains details of a device and its serial port
    --------------------------------------------------------------------------
    returns : string
    --------------------------------------------------------------------------
    """
    message = None

    if dev.model in ('hmp4040', '2410', '2614b'):
        command_string = lexicon.power(dev.model, 'get current limit', channel=dev.channel)
        response = common.atomic_send_command_read_response(pipeline, ser, dev, command_string)
        with contextlib.suppress(ValueError):
            message = f'{common.si_prefix(response)}A'

    return message


def simple_output_status(ser, pipeline, dev, tcolours):
    """
    Return the on/off output status of the given power supply channel.

    Values returned in variable output:

    +------------------+-----------+---------------+---------------+
    | dev.manufacturer | dev.model | output OFF    | output ON     |
    +------------------+-----------+---------------+---------------+
    | 'agilent'        | 'e3634a'  | '0'           | '1'           |
    | 'agilent'        | 'e3647a'  | '0'           | '1'           |
    | 'hameg'          | 'hmp4040' | '0'           | '1'           |
    | 'iseg'           | 'shq'     | '=OFF'        | '=ON'         |
    | 'keithley'       | '2410'    | '0'           | '1'           |
    | 'keithley'       | '2614b'   | '0.00000e+00' | '1.00000e+00' |
    +------------------+-----------+---------------+---------------+

    --------------------------------------------------------------------------
    args
        ser : serial.Serial
            reference for serial port
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        dev : instance of class Channel
            contains details of a device and its serial port
        tcolours : class
            contains ANSI colour escape sequences
    --------------------------------------------------------------------------
    returns : string
    --------------------------------------------------------------------------
    """
    command_string = lexicon.power(dev.model, 'check output', channel=dev.channel)
    output = common.atomic_send_command_read_response(pipeline, ser, dev, command_string)

    status_categories = {
        True: f'{tcolours.BOLD}on{tcolours.ENDC}',
        False: 'off',
        None: 'could not be determined'}

    channel_on = None

    if dev.manufacturer == 'iseg':
        if any(x in output for x in ['=ON', '=OFF']):
            channel_on = '=ON' in output

    elif dev.manufacturer in {'agilent', 'hameg', 'keithley'}:
        try:
            outval = int(float(output))
        except ValueError:
            pass
        else:
            if outval in {0, 1}:
                channel_on = outval == 1

    return status_categories[channel_on]


def simple_interlock_status(ser, pipeline, dev, tcolours):
    """
    Check status of power supply interlock.

    This is per-PSU on Keithley, per-channel on ISEG.

    --------------------------------------------------------------------------
    args
        ser : serial.Serial
            reference for serial port
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        dev : instance of class Channel
            contains details of a device and its serial port
        tcolours : class
            contains ANSI colour escape sequences
    --------------------------------------------------------------------------
    returns : string
    --------------------------------------------------------------------------
    """
    assert dev.manufacturer not in {'agilent', 'hameg'},\
        'function not callable for Agilent or Hameg PSU'

    status_categories = {
        True: 'active',
        False: f'{tcolours.BOLD}inactive{tcolours.ENDC}',
        None: 'could not be determined'}

    interlock_set = None

    command_string = lexicon.power(dev.model, 'check interlock', channel=dev.channel)
    register = common.atomic_send_command_read_response(pipeline, ser, dev, command_string)

    if dev.manufacturer == 'iseg':
        # Interlocks are per-channel for ISEG SHQ.
        interlock_set = '=INH' in register
    else:
        # Interlocks are per-PSU for Keithley.
        try:
            regval = int(float(register))
        except ValueError:
            pass
        else:
            if dev.model == '2614b':
                # bit 11 (status.measurement.INTERLOCK) p.11-280 (648)
                # Without hardware interlock: '0.00000e+00'
                # with hardware interlock: '2.04800e+03'
                interlock_set = regval & 2048 == 0
            elif dev.model == '2410':
                if regval in {0, 1}:
                    # p.18-9 (389)
                    # returns '0' (disabled - no restrictions) or '1' (enabled)
                    interlock_set = register == 0

    return status_categories[interlock_set]


def simple_polarity_status(ser, pipeline, dev, tcolours):
    """
    Returned string from 'check module status' command appears to be base 10.

    e.g.
        inhibit is bit 5 (0=inactive, 1=active),
        polarity is bit 2 (0=negative, 1=positive)

        For channel 1, with inhibit active (terminator removed from front
        panel inhibit socket) and the polarity switch set to POS on the
        rear panel, the 'S1' command returns '036'

    e.g.
        polarity set to positive: '004', negative: '000'

    --------------------------------------------------------------------------
    args
        ser : serial.Serial
            reference for serial port
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        dev : instance of class Channel
            contains details of a device and its serial port
        tcolours : class
            contains ANSI colour escape sequences
    --------------------------------------------------------------------------
    returns : string
    --------------------------------------------------------------------------
    """
    assert dev.model == 'shq', 'function only callable for ISEG SHQ PSU'

    status_categories = {
        True: 'positive',
        False: 'negative',
        None: f'{tcolours.BG_RED}{tcolours.FG_WHITE}could not be determined{tcolours.ENDC}'}

    polarity_positive = None

    command_string = lexicon.power(dev.model, 'check module status', channel=dev.channel)
    register = common.atomic_send_command_read_response(pipeline, ser, dev, command_string)

    with contextlib.suppress(ValueError):
        polarity_positive = int(register) & 4 != 0

    return status_categories[polarity_positive]


def power_supply_channel_status(settings, pipeline, psus, channels, tcolours):
    """
    Establishes RS232 communications with power supplies (as required).
    Checks status of channel outputs and interlocks (inhibits).

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        psus : dict
            {port: ({port_config}, device_type, device_serial_number), ...}
            contents of the cache filtered by hvpsu category
        channels : list
            contains instances of class Channel, one for each
            power supply channel
        tcolours : class
            contains ANSI colour escape sequences
    --------------------------------------------------------------------------
    returns
        channels : list
            no explicit return, mutable type amended in place
    --------------------------------------------------------------------------
    """
    port_used = collections.defaultdict(int)

    # this function call may change the value of variable channels
    # and returns a dict containing a mapping of ports to the serial ports
    # which have been left open
    spd = common.check_ports_accessible(psus, channels, close_after_check=False)

    for dev in channels:
        read_values = []

        ######################################################################
        # identify channel
        ######################################################################

        if not settings['filter']:
            print()
            print(dev)

        # specify already opened serial port
        ser = spd[dev.port]

        # try to ensure consistent state
        # clear FTDI output buffer state before sending
        ser.reset_output_buffer()

        # clear PSU state
        if dev.model == '2614b':
            command_string = lexicon.power(dev.model, 'terminator only',
                                           channel=dev.channel)
            common.send_command(pipeline, ser, dev, command_string)

        # clear FTDI input buffer state
        ser.reset_input_buffer()

        # arbitrary settle time before proceeding
        time.sleep(0.0)

        # ensure serial port communication with PSU
        # this is for ISEG SHQ only, on a per-PSU basis
        if not port_used[dev.port]:
            common.synchronise_psu(ser, pipeline, dev)

        ##################################################################
        # determine interlock status
        ##################################################################

        if not settings['filter']:
            # this is per-PSU on Keithley, per-channel on ISEG
            # still report it per-channel either way.
            if dev.manufacturer not in {'agilent', 'hameg'}:
                print(f'interlock status: {simple_interlock_status(ser, pipeline, dev, tcolours)}')

        ##################################################################
        # determine output status
        ##################################################################

        out_stat = simple_output_status(ser, pipeline, dev, tcolours)

        if not settings['filter']:
            print(f'output status: {out_stat}')
            set_volt = read_psu_set_voltage(pipeline, ser, dev)
            if set_volt is None:
                sv_text = 'could not be determined'
            else:
                sv_text = f'{common.si_prefix(set_volt)}V'
            read_values.append(f'set voltage: {sv_text}')

        ##################################################################
        # determine current limit
        ##################################################################

        current_limit = simple_current_limit(ser, pipeline, dev)
        if current_limit is not None and not settings['filter']:
            print(f'current limit: {current_limit}')

        ##################################################################
        # report measured values
        ##################################################################

        if 'on' in out_stat:
            mvolt, mcurr = read_psu_vi(pipeline, ser, dev)
            if mvolt is None:
                mvi_text = 'could not be determined'
            else:
                if not settings['filter']:
                    mvi_text = f'{common.si_prefix(mvolt)}V, {common.si_prefix(mcurr)}A'
                else:
                    if settings['mi']:
                        print(mcurr)
                    elif settings['mv']:
                        print(mvolt)
                    else:
                        print('None')

            if not settings['filter']:
                read_values.append(f'measured: {mvi_text}')

        else:
            print('None')

        ##################################################################
        # determine polarity status
        ##################################################################

        if not settings['filter']:
            if dev.manufacturer == 'iseg':
                print(f'polarity: {simple_polarity_status(ser, pipeline, dev, tcolours)}')

        ##################################################################
        # determine constant voltage/current operation
        ##################################################################

        if not settings['filter']:
            if dev.manufacturer == 'hameg':
                print(f'mode: {hmp4040_constant_voltage_current_status(ser, pipeline, dev, tcolours)}')

        ser.reset_input_buffer()
        ser.reset_output_buffer()

        if not settings['filter']:
            print(', '.join(read_values))

        port_used[dev.port] += 1

    # close all the serial ports left open by the call above to
    # common.check_ports_accessible()
    for serial_port in spd.values():
        serial_port.close()


def read_psu_vi(pipeline, ser, dev):
    """
    read measured voltage and current from PSU

    --------------------------------------------------------------------------
    args
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        ser : serial.Serial
            reference for serial port
        dev : instance of class Channel
            contains details of a device and its serial port
    --------------------------------------------------------------------------
    returns
        dev : instance of class Channel
            contains details of a device and its serial port
        readings : list
            [(float, float)]
    --------------------------------------------------------------------------
    """
    if dev.manufacturer in {'keithley', 'iseg'}:
        volt, curr = common.read_psu_measured_vi(pipeline, ser, dev)
    elif dev.manufacturer in {'agilent', 'hameg', 'hewlett-packard'}:
        if dev.model == 'e3634a':
            command_string = lexicon.power(dev.model, 'set remote')
            common.send_command(pipeline, ser, dev, command_string)

        command_string = lexicon.power(dev.model, 'read voltage', channel=dev.channel)
        measured_voltage = common.atomic_send_command_read_response(pipeline, ser,
                                                                    dev, command_string)

        command_string = lexicon.power(dev.model, 'read current', channel=dev.channel)
        measured_current = common.atomic_send_command_read_response(pipeline, ser,
                                                                    dev, command_string)
        try:
            volt = float(measured_voltage)
            curr = float(measured_current)
        except ValueError:
            volt = curr = None

        # ignore readings if output is off
        if dev.manufacturer == 'hameg':
            command_string = lexicon.power(dev.model, 'check output', channel=dev.channel)
            output_status = common.atomic_send_command_read_response(pipeline, ser,
                                                                     dev, command_string)
            if output_status == '0':
                volt = curr = None

    else:
        volt = curr = None

    return volt, curr


def read_psu_set_voltage(pipeline, ser, dev):
    """
    Read the voltage that the PSU has been asked to output (rather than the
    actual instantaneous voltage measured at the output terminals).

    --------------------------------------------------------------------------
    args
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        ser : serial.Serial
            reference for serial port
        dev : instance of class Channel
            contains details of a device and its serial port
    --------------------------------------------------------------------------
    returns
        set_volt : float or None
            float if the value could be read or None if it could not
    --------------------------------------------------------------------------
    """
    set_volt = None

    command_string = lexicon.power(dev.model, 'read set voltage', channel=dev.channel)
    local_buffer = common.atomic_send_command_read_response(pipeline, ser, dev, command_string)

    if dev.manufacturer == 'iseg' and dev.model == 'shq':
        set_volt = common.iseg_value_to_float(local_buffer)

    elif dev.manufacturer == 'keithley' and dev.model in {'2410', '2614b'}:
        if dev.model == '2410':
            # e.g. '-5.000000E+00,-1.005998E-10,+9.910000E+37,+1.185742E+04,+2.150800E+04'
            item = next(iter(local_buffer.split(',')))
        else:
            # e.g. '-5.00000e+00'
            item = local_buffer

        with contextlib.suppress(TypeError, ValueError):
            set_volt = float(item)

    elif dev.manufacturer == 'hameg':
        with contextlib.suppress(ValueError):
            set_volt = float(local_buffer)

    return set_volt


##############################################################################
# main
##############################################################################

def main():
    """
    Instantaneous snapshot of all power supply channels from devices
    connected by FTDI USB to RS232 adaptors.
    """
    settings = {
        'alias': None,
        'channel': None,
        'manufacturer': None,
        'model': None,
        'port': None,
        'serial': None,
        'time': None,
        }

    check_arguments(settings)

    ##########################################################################
    # read cache
    ##########################################################################

    psus = common.cache_read(['hvpsu', 'lvpsu'], quiet=settings['filter'])
    channels = common.ports_to_channels(settings, psus)

    if settings['filter'] and not common.unique(settings, psus, channels):
        print('could not identify a single power supply channel')
        print('check use of --manufacturer, --model, --serial, --port, and --channel')
        sys.exit()

    ##########################################################################
    # set up resources for threads
    ##########################################################################

    class Production:
        """
        Queues and locks to support threaded operation.

        RS232 will not tolerate concurrent access. portaccess is used to
        prevent more than one thread trying to write to the same RS232 port at
        the same time for multi-channel power supplies. For simplicity, locks
        are created for all power supplies, even if they only have a single
        channel.
        """
        portaccess = {port: threading.Lock()
                      for port in {channel.port for channel in channels}}

    pipeline = Production()

    ##########################################################################
    # Check status of outputs and interlock (inhibit) on all power supplies
    ##########################################################################

    power_supply_channel_status(
        settings, pipeline, psus, channels, common.ANSIColours
    )


##############################################################################
if __name__ == '__main__':
    main()



"""
Common functions for number sequence generation.

callable functions from this file:

    test_run
    to_original
    to_start
"""

import collections
import functools
import operator as op
import os
import sys

from mmcb import common


##############################################################################
# number sequence generation
##############################################################################

def _sequence(start, stop, step, variablestep, initial, final, bespoke_sequence_lut=None):
    """
    Generate a number sequence with constant or variable step sizes.

    Start and end points may be floating point values, all other generated
    values are integers aligned with the appropriate step size.

    --------------------------------------------------------------------------
    args
        start : numeric
            start value of the number sequence
        stop : numeric
            stop value of the number sequence
        step : int
            step size to which steps are aligned to
        variablestep : bool
            True if variable step sizes will be generated by this function,
            False otherwise
        initial : bool
            include initial value (start) in number sequence
        final : bool
            include final value (stop) in number sequence
        bespoke_sequence_lut : dict if value is supplied, or None default value
            A lookup table for determining step sizes and signs.
            This value is not used by this function, but is passed directly
            to _stepped_step() via a partial function. See _stepped_step()
            documentation for more details.
    --------------------------------------------------------------------------
    yields : series of numeric values
        first and last values may be float or int, all others will be int
    --------------------------------------------------------------------------
    """
    sign = 1 if start < stop else -1
    _step = functools.partial(_stepped_step,
                              variablestep=variablestep, step=step, sign=sign,
                              bespoke_sequence_lut=bespoke_sequence_lut)

    start_rounded = common.round_safely(start, stop, _step(start), first=True)
    stop_rounded = common.round_safely(start, stop, _step(stop), first=False)
    aligned_start = start == start_rounded
    aligned_stop = stop == stop_rounded

    # yield non-aligned initial value
    if initial and not aligned_start:
        yield start

    # yield aligned number sequence
    num = start_rounded
    if not initial and aligned_start:
        num += _step(start_rounded)
    compare_value = (stop_rounded if not final and aligned_stop
                     else stop_rounded + _step(stop_rounded))
    comparison = op.lt if sign == 1 else op.gt
    while comparison(num, compare_value):
        yield num
        num += _step(num)

    # yield non-aligned final value
    if final and not aligned_stop:
        yield stop


# wrappers for _sequence
to_start = functools.partial(_sequence, variablestep=False, initial=False, final=False)
test_run = functools.partial(_sequence, initial=True, final=True)
to_original = functools.partial(_sequence, variablestep=False, initial=False, final=True)


def _stepped_step(value, variablestep, step, sign, bespoke_sequence_lut):
    """
    Calculate the correct step size for the given value given the band it lies
    within, and the direction of travel of the number sequence.

       band  step size
     0 - 10          1
    10 - 50          5
       > 50         10

    op.lt    less than
    op.gt    greater than
    op.le    less than or equal to
    op.ge    greater than or equal to

    --------------------------------------------------------------------------
    args
        value : numeric
            either positive or negative
        variablestep : bool
            True if variable step sizes will be generated by this function,
            False otherwise
        step : int
            step size (positive value only in the context of this script)
        sign : int
            either -1 or 1 (signifies the direction of travel of the number
            sequence)
        bespoke_sequence_lut : dict
            A lookup table for determining step sizes and signs.
            See unit tests in test_sequence.py for more information.
            e.g.
            {(-1, False): [[10, op.le, 1], [50, op.le, 5], [70, op.le, 10],
                           [80, op.le, 1], [None, None, 10]],
             (-1, True): [[10, op.lt, 1], [50, op.lt, 5], [70, op.lt, 10],
                          [80, op.lt, 1], [None, None, 10]],
             (1, False): [[80, op.ge, 10], [70, op.ge, 1], [50, op.ge, 10],
                          [10, op.ge, 5], [None, None, 1]],
             (1, True): [[80, op.gt, 10], [70, op.gt, 1], [50, op.gt, 10],
                         [10, op.gt, 5], [None, None, 1]]}
    --------------------------------------------------------------------------
    returns : int
        the appropriate step size for the number range that value lies within
    --------------------------------------------------------------------------
    """
    if variablestep:
        # the order of the items in the value list is important
        test = {(-1, False): [[10, op.le,  1], [50, op.le,  5], [None, None, 10]],
                (-1,  True): [[10, op.lt,  1], [50, op.lt,  5], [None, None, 10]],
                ( 1, False): [[50, op.ge, 10], [10, op.ge,  5], [None, None,  1]],
                ( 1,  True): [[50, op.gt, 10], [10, op.gt,  5], [None, None,  1]]}
        if bespoke_sequence_lut is not None:
            test = bespoke_sequence_lut

        direction = value < 0
        value = abs(value)
        for boundary, comparison, stepsize in test[(sign, direction)]:
            if boundary is not None and comparison(value, boundary):
                break
    else:
        stepsize = abs(step)

    return stepsize * sign


##############################################################################
# allow the use of a user-defined file containing ranges and step values
##############################################################################

def _load_test_profile(filename):
    """
    Load a file containing a series of ranges and step sizes that will define
    the voltages that will be used during the IV test.

    Comments starting with '#' are supported. On any line if a '#' is found,
    it and any text to the right are discarded. Blank lines are allowed.

    Valid ranges are either (1) a single value that defines the default step
    size over the entire range from 0V to the maximum test voltage, and
    (2) three values (start, stop, step).

    If errors are encountered reading the file, continue reading the file
    anyway so the user receives all error messages at once.

    --------------------------------------------------------------------------
    args
        filename : string
            name of file containing range profile
    --------------------------------------------------------------------------
    returns
        profile : list or None if file contained invalid line(s)
            [[float, int], ...]
            e.g. a file containing:
                # default step size
                10
                # list all ranges required: start, stop, step
                0, 10, 1
                10, 50, 5
            would return:
                [[10], [0, 10, 1], [10, 50, 5]]
    --------------------------------------------------------------------------
    """
    error = False
    profile = []

    if os.path.exists(filename):
        with open(filename, 'r', encoding='utf-8') as csvfile:
            for row in csvfile:
                no_comment = row.split('#')[0].strip()

                row = [field.strip() for field in no_comment.split(',')]
                rowlen = len(row)

                if rowlen in {1, 3}:
                    if all(bool(x) for x in row):
                        profile.append([int(x) for x in row])
                else:
                    print(f'--range: malformed line: {no_comment}')
                    error = True
                    break
    else:
        message = f'--range: file {filename} could not be read from'
        print(message)

    return None if error else profile


def _check_ranges(ranges):
    """
    Basic checks on received ranges.

    --------------------------------------------------------------------------
    args
        ranges : list
            name of file containing range profile
            e.g. [[10], [0, 10, 1], [10, 50, 5]]
    --------------------------------------------------------------------------
    returns
        ranges : no explicit return, mutable type amended in place
            the line is sorted so the default step size is first in the list
                e.g. [[10], [0, 10, 1], [10, 50, 5]]
    --------------------------------------------------------------------------
    """
    counts = collections.Counter(len(x) for x in ranges)
    default_step_size_count = counts[1]

    if default_step_size_count == 0:
        print('--range: no default step size specified')
        sys.exit()
    elif default_step_size_count > 1:
        # don't do anything clever, let the user fix the problem in their file
        defaults = [x[0] for x in ranges if len(x) == 1]
        details = ', '.join(f'{x}' for x in defaults)
        print(f'--range: multiple default voltage step sizes provided: {details}')
        sys.exit()


def _clean(maxval, item):
    """
    --------------------------------------------------------------------------
    args
        maxval : int
            the maximum specified test voltage
        item : list
            e.g. [0, 10, 1]
    --------------------------------------------------------------------------
    returns
        item : no explicit return, mutable type amended in place
            the line is sorted so the default step size is first in the list
                e.g. [0, 10, 1]
    --------------------------------------------------------------------------
    """
    # make sure values are positive
    item[:] = [abs(i) for i in item]

    if len(item) == 3:
        # ensure elements are in the correct order for range(start, stop, step)
        if item[0] > item[1]:
            item[0], item[1] = item[1], item[0]

        # make sure no values will be generated beyond maxval
        if item[1] > abs(maxval):
            item[1] = abs(maxval)


def _entire_sequence(maxval, maxstep, ranges):
    """
    Generate a sequence of voltages with different step sizes spanning the
    full voltage range.

    --------------------------------------------------------------------------
    args
        maxval : int
            the maximum specified test voltage
        maxstep : int
            the maximum step size
        ranges : list
            e.g. [[10], [0, 10, 1], [10, 50, 5]]
    --------------------------------------------------------------------------
    yields : series of positive ints
        These represent the full series of voltage points spanning the range
        from 0V to the maximum test voltage.
            e.g. [  0,   1,   2,   3,   4,   5,   6,   7,   8,   9,
                   10,  15,  20,  25,  30,  35,  40,  45,  50,  60,
                   70,  80,  90, 100]
    --------------------------------------------------------------------------
    """
    # position low-length items first to make sure the default step size is
    # processed first.
    ranges.sort(key=len)

    seq = set()

    for ran in ranges:
        _clean(maxval, ran)

        try:
            start, stop, step = ran
        except ValueError:
            # Populate range with default step size.
            # Make sure we go one step beyond the end to ensure final values
            # that are not aligned to the step size will not be missed
            seq |= set(range(0, abs(maxval) + abs(maxstep), ran[0]))
            continue

        # all remove numbers inside the current range
        seq -= set(range(start, stop + 1))

        # fill set with specified range of numbers
        seq |= set(range(start, stop + 1, step))

    # low to high order
    seqc = sorted(seq)

    # reject step sequences that may damage the device under test
    diffs = [s - f for f, s in zip(seqc, seqc[1:])]
    if any(d > abs(maxstep) for d in diffs):
        print(f'--range: steps are too large {max(diffs)}')
        sys.exit()

    # deliver results
    for voltage in seqc:
        yield voltage


def _sequence_to_ranges(numseq):
    """
    Takes a number sequence and turns it into a list of ranges of equal
    consecutive step sizes that can be used to reproduce the number
    sequence.

    --------------------------------------------------------------------------
    args
        numseq : list of positive ints
            Represent the full series of voltage points spanning the range
            from 0V to the maximum test voltage. Values will always be
            positive.
            e.g. [  0,   1,   2,   3,   4,   5,   6,   7,   8,   9,
                   10,  15,  20,  25,  30,  35,  40,  45,  50,  60,
                   70,  80,  90, 100]
    --------------------------------------------------------------------------
    returns
        ranges : list
            e.g. [[0, 10, 1], [10, 50, 5], [50, 100, 10]]
    --------------------------------------------------------------------------
    """
    ranges = []

    if not numseq:
        return ranges

    last_diff = start = second = None

    for first, second in zip(numseq, numseq[1:]):
        if last_diff is None:
            last_diff = abs(second - first)
            start = first
        else:
            this_diff = abs(second - first)
            if this_diff == last_diff:
                continue

            ranges.append([start, first, last_diff])
            start = first
            last_diff = this_diff

    if second is not None:
        ranges.append([start, second, last_diff])

    return ranges


def _ranges_to_lut(ranges):
    """
    From a list of ranges that completely covers the whole number sequence,
    generate a lookup table dictionary usable by _stepped_step().

    --------------------------------------------------------------------------
    args
        ranges : list
            e.g. [[0, 10, 1], [10, 50, 5], [50, 100, 10]]
    --------------------------------------------------------------------------
    returns
        lut : dict
            e.g.
            {(-1, False): [[10, op.le,  1], [50, op.le, 5], [None, None, 10]],
             (-1,  True): [[10, op.lt,  1], [50, op.lt, 5], [None, None, 10]],
             ( 1, False): [[50, op.ge, 10], [10, op.ge, 5], [None, None,  1]],
             ( 1,  True): [[50, op.gt, 10], [10, op.gt, 5], [None, None,  1]]}
    --------------------------------------------------------------------------
    """
    first = 0
    lut = collections.defaultdict(list)

    # key (sign, direction (value < 0))
    config = {
        (-1, False): op.le,
        (-1, True): op.lt,
        (1, False): op.ge,
        (1, True): op.gt,
    }

    for sign_direction, comp in config.items():
        sign, _direction = sign_direction

        for index, (norm, reve) in enumerate(zip(ranges, reversed(ranges))):
            nstart, _nstop, nstep = norm
            _rstart, rstop, rstep = reve

            if index == first:
                if sign == -1:
                    lut[sign_direction].append([None, None, rstep])
                else:
                    lut[sign_direction].append([None, None, nstep])
            else:
                if sign == -1:
                    lut[sign_direction].append([rstop, comp, rstep])
                else:
                    lut[sign_direction].append([nstart, comp, nstep])

    # the order of the items in the value list is important
    # sign == -1: from closest to zero to furthest from zero
    # sign ==  1: from furthest from zero to closest to zero
    for key, value in lut.items():
        sign, _direction = key
        if sign == 1:
            value.sort(key=lambda x: 0 if x[0] is None else x[0], reverse=True)
        else:
            value.sort(key=lambda x: 10000 if x[0] is None else x[0])

    return dict(lut)


def read_user_range_step_file(settings, filename):
    """
    Read ranges from file, generate a lookup table suitable for
    _stepped_step.

    --------------------------------------------------------------------------
    args
        settings : dictionary
            contains core information about the test environment
        filename : string
            filename with extension
    --------------------------------------------------------------------------
    returns
        settings : dict
            no explicit return, mutable type amended in place
    --------------------------------------------------------------------------
    """
    maxval = settings['voltage']
    maxstep = 10

    # read file
    ranges = _load_test_profile(filename)
    if ranges is None:
        sys.exit()

    # make sure it's not broken
    _check_ranges(ranges)

    # generate full number sequence
    full_number_sequence = list(_entire_sequence(maxval, maxstep, ranges))

    # Now reduce the full number sequence to a list of ranges that fully define
    # the voltage range.
    all_ranges = _sequence_to_ranges(full_number_sequence)

    # now transform the all ranges list to something that can be used by
    # _stepped_step to manage its number sequences
    lut = _ranges_to_lut(all_ranges)

    settings['bespoke_sequence_lut'] = lut
